/*
 * This file is part of decoding
 * Copyright 2012-2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <decoding/decoding.h>

#define RING_NAME fld
#define PRIME 257
#include <decoding/rings/gfp_word.c>
#include <decoding/algos.c>

void vec_print_list(FILE *out,fld_vec m,fld_vec *a,uma na) {
  uma i;
  for( i = 0 ; i < na ; i++ ) {
    fld_vec_print(out,a[i]);
    fprintf(stderr," at distance %ld\n",
                   (long)fld_vec_hamming_distance(m,a[i]));
  }
}

void die(fld_vec m,fld_vec *a,uma na) {
  fprintf(stderr,"Error: cannot find message:\n");
  fld_vec_print(stderr,m); fputc('\n',stderr);
  if ( na > 0 ) {
    fprintf(stderr,"List returned:\n");
    vec_print_list(stderr,m,a,na);
  }
  exit(-1);
}

void sudan(uma n,uma k,uma tau,uma sec) {
  uma i,nc,count;
  fld_vec m,y,e,*c;
  fld_rs_code rs;
  clock_t start,end,delta;

  fld_vec_inits(m,y,e,NULL);
  fld_rs_code_init_with_support(rs,n,k);
  /* rad = sudan_max_errors(n,k); */
  /* tau = (int)((double)percent * (double)rad / 100.0); */
  delta = (clock_t)(sec * CLOCKS_PER_SEC);
  count = 0;
  start = clock();
  while(1) {
    fld_rs_code_random_codeword(m,rs);
    fld_vec_random_hamming_weight(e,n,tau);
    fld_vec_add(y,m,e);
    nc = fld_rs_code_sudan_koetter(&c,rs,y,tau);
    if ( nc == 0 ) die(m,c,0);
    for( i = 0 ; i < nc ; i++ ) {
      if ( fld_vec_cmp(c[i],m) == 0 ) goto sudan_good;
    }
    die(m,c,nc);
    sudan_good:;
    fld_vec_ptr_clear(c,nc);
    end = clock();
    count++;
    if ( end > start + delta ) break;
  }
  fld_rs_code_clear(rs);
  fld_vec_clears(m,y,e,NULL);
  printf("%ld  %ld  %ld  %f\n",n,k,tau,
         ((double)(end - start) / CLOCKS_PER_SEC / count));
}

void guruswami_sudan(uma n,uma k,uma tau,uma sec) {
  uma i,nc,count;
  fld_vec m,y,e,*c;
  fld_rs_code rs;
  clock_t start,delta,end;

  fld_vec_inits(m,y,e,NULL);
  fld_rs_code_init_with_support(rs,n,k);
  /* rad = guruswami_sudan_max_errors(n,k); */
  /* tau = (int)((double)percent * (double)rad / 100.0); */
  delta = (clock_t)(sec * CLOCKS_PER_SEC);
  count = 0;
  start = clock();
  while(1) {
    fld_rs_code_random_codeword(m,rs);
    fld_vec_random_hamming_weight(e,n,tau);
    fld_vec_add(y,m,e);
    nc = fld_rs_code_guruswami_sudan_koetter(&c,rs,y,tau);
    if ( nc == 0 ) die(m,c,0);
    for( i = 0 ; i < nc ; i++ ) {
      if ( fld_vec_cmp(c[i],m) == 0 ) goto guruswami_sudan_good;
    }
    die(m,c,nc);
    guruswami_sudan_good:;
    fld_vec_ptr_clear(c,nc);\
    end = clock();
    count++;
    if ( end > start + delta ) break;
  }
  fld_rs_code_clear(rs);
  fld_vec_clears(m,y,e,NULL);
  printf("%ld  %ld  %ld  %f\n",n,k,tau,
         ((double)(end - start) / CLOCKS_PER_SEC / count));
}

#define N 256
#define K 64

int main(void) {
  uma i,min,max;
  srand(time(0));
  fld_ring_init();
  /*
  min = (N - K) / 2;
  max = sudan_max_errors(N,K);
  for( i = min ; i <= max ; i++ ) {
    sudan(N,K,i,2);
  }
  fputs("\n\n",stdout);
  */
  min = (N - K) / 2;
  max = guruswami_sudan_max_errors(N,K) - 5;
  for( i = min ; i <= max ; i++ ) {
    guruswami_sudan(N,K,i,2);
  }
  return 0;
}

