/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
The `uma` type (Unaligned Memory Access) shall be used everywhere
you need indices. In particular, all function in decoding returning
polynomial degrees, valuations, number of elements within a certain
set or list are all of type `uma`. It usually corresponds to `long`,
but do not assume that. You can safely assume that every `int` can be
casted to `uma` without problem. When using `decoding` remember to
use `uma` and *not* `int` for array indices.

The `crit_*` family of funtions indicates a *critical* zone in the
code where all the arithmetic operations involving `uma`'s *must*
be performed correctly. For example, suppose that you have to
multiply two polynomials of degree `d1` and `d2`. Then you shall
use

``
uma d3;
d3 = crit_add(d1,d2);
``

to compute the degree of the product of the two polynomials.
If the sum of `d1` and `d2` does not hold within a machine
word the `SIGABRT` signal is raised, otherwise `d3` will
be equal to `d1 + d2`. All arithmetic operations have their
`crit_` counterpart.
*/

void crit_common_check(uma a,uma b,const char *file,long line) {
  if ( a < 0 ) {
    fprintf(stderr,"decoding: a must be non negative\n"
                   "at %s:%ld\n",file,line);
    abort();
  }
  if ( b < 0 ) {
    fprintf(stderr,"decoding: b must be non negative\n"
                   "at %s:%ld\n",file,line);
    abort();
  }
}

/**
> `uma crit_add(a,b);`
  Return `a + b` if it holds within a machine.
  If `a` or `b` is negative or if it overflows the `SIGABRT`
  signal is raised.
*/
#define crit_add(a,b) __crit_add((a),(b),__FILE__,__LINE__)
uma __crit_add(uma a,uma b,const char *file,long line) {
  crit_common_check(a,b,file,line);
  if ( b > (uma_max - a) ) {
    fprintf(stderr,"decoding: overflow: %ld + %ld > %ld\n"
                   "at %s:%ld\n",
                   __CAST(long,a),__CAST(long,b),__CAST(long,uma_max),
                   file,line);
    abort();
  }
  return a + b;
}

/**
> `uma crit_sub(a,b);`
  Return `a - b` if it holds within a machine.
  If `a` or `b` is negative or if it overflows the `SIGABRT`
  signal is raised.
*/
#define crit_sub(a,b) __crit_sub((a),(b),__FILE__,__LINE__)
uma __crit_sub(uma a,uma b,const char *file,long line) {
  crit_common_check(a,b,file,line);
  if ( b > a ) {
    fprintf(stderr,"decoding: underflow: %ld - %ld < 0\n"
                   "at %s:%ld\n",
                   __CAST(long,a),__CAST(long,b),file,line);
    abort();
  }
  return a - b;
}

/**
> `uma crit_mul(a,b);`
  Return `a * b` if it holds within a machine.
  If `a` or `b` is negative or if it overflows the `SIGABRT`
  signal is raised.
*/
#define crit_mul(a,b) __crit_mul((a),(b),__FILE__,__LINE__)
uma __crit_mul(uma a,uma b,const char *file,long line) {
  crit_common_check(a,b,file,line);
  if ( b > ((uma_max / a) + 1) ) {
    fprintf(stderr,"decoding: overflow: %ld * %ld > %ld\n"
                   "at %s:%ld\n",
                   __CAST(long,a),__CAST(long,b),__CAST(long,uma_max),
                   file,line);
    abort();
  }
  return a * b;
}

/**
> `uma crit_div(a,b);`
  Return `a / b`.
  If `a` or `b` is negative or if it overflows the `SIGABRT`
  signal is raised.
*/
#define crit_div(a,b) __crit_div((a),(b),__FILE__,__LINE__)
uma __crit_div(uma a,uma b,const char *file,long line) {
  crit_common_check(a,b,file,line);
  return a / b;
}

