/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Root-finding algorithm for Guruswami-Sudan.
   From: Polynomial root finding over local rings and
         application to error correcting codes.
         Jeremy Berthomieu, Gregoire Lecerf, Guillaume Quintin.

   In fact, Roth and Ruckenstein did nothing more than Newton
   substitution. Our paper prooves that their approach is
   polynomial even for ``non Guruswami-Sudan" polynomials.

   Here is a mix of the naive version of our root-finding
   algorithm and Roth & Ruckenstein's.
*/

struct ll_vec_lili *ll_dbpol_Xroots_rec(R *a,uma nax,uma nay,
                                        uma m,uma w,uma n)
{
  R *z0,*q;
  uma i,v,na,ai,nz0,nq,nqy,nret,nroots;
  struct ll_vec_lili *z1,*ret,*node,*nn,*last;
  
  ret = NULL;
  last = NULL;
  na = nax * nay;
  for( ai = 0 ; ai < na && ll_vec_iszero(&a[ai],nax) ; ai += nax,nay-- );
  if ( ai >= na ) return NULL;
  nz0 = m - 1;
  z0 = ll_vec_malloc_init(nz0);
  ll_upol_roots(z0,&nroots,&a[ai],m);
  if ( nroots == 0 ) {
    ret = NULL;
    goto ll_dbpol_Xroots_rec_free_z0;
  }
  if ( w >= n - 1 ) { /* found lasts coefficients */
    ret = ll_vec_lili_node_malloc(1);
    R_copy(ret->a[0],z0[0]);
    last = ret;
    for( i = 1 ; i < nroots ; i++ ) {
      nn = ll_vec_lili_node_malloc(1);
      R_copy(nn->a[0],z0[i]);
      ll_vec_lili_node_attach_after(last,nn);
      last = nn;
    }
    goto ll_dbpol_Xroots_rec_free_z0;
  }
  else { /* recursive call to get the queues */
    nq = nax * (nay + nax);
    q = ll_vec_malloc_init(nq);
    nret = n - w;
    for( i = 0 ; i < nroots ; i++ ) {
      ll_dbpol_shift_in_X(q,&a[ai],nax,nay,z0[i]);
      for( v = 0 ; v < nax && R_iszero(q[v]) ; v++ );
      if ( v >= nax ) { /* should never happen */
        ret = NULL;
        goto ll_dbpol_Xroots_rec_free_q;
      }
      ll_dbpol_replace_X_by_XYn(q,&nqy,q,nax,nay,1);
      z1 = ll_dbpol_Xroots_rec(q,nax,nqy,v + 1,w + 1,n);
      if ( z1 ) {
        for( node = z1 ; node ; node = node->next ) {
          if ( last ) {
            nn = ll_vec_lili_node_malloc(nret);
            R_copy(nn->a[0],z0[i]);
            ll_vec_copy(&(nn->a[1]),node->a,nret - 1);
            ll_vec_lili_node_attach_after(last,nn);
            last = nn;
          }
          else {
            ret = ll_vec_lili_node_malloc(nret);
            R_copy(ret->a[0],z0[i]);
            ll_vec_copy(&(ret->a[1]),z1->a,nret - 1);
            last = ret;
          }
        }
        ll_vec_lili_list_free(z1);
      }
    }
    goto ll_dbpol_Xroots_rec_free_q;
  }

ll_dbpol_Xroots_rec_free_q:
  ll_vec_free(q,nq);

ll_dbpol_Xroots_rec_free_z0:
  ll_vec_free(z0,nz0);

  return ret;
}

/**
> `struct ll_vec_lili *ll_dbpol_Xroots(R *a,uma nax,uma nay,uma n);`
  Return a linked list containing the polynomials `f` such that
  `a(f(X),X) = 0 mod X^n` where `a` has size (`nax`,`nay`).
  The algorithm is a mix between Rock and Ruckenstein's and
  Berthomieu, Lecerf and Quintin's.
*/

struct ll_vec_lili *ll_dbpol_Xroots(R *a,uma nax,uma nay,uma n) {
#ifdef MY_DBPOL_XROOTS
  return MY_DBPOL_XROOTS(a,nax,nay,nax,n);
#else
  return ll_dbpol_Xroots_rec(a,nax,nay,nax,0,n);
#endif
}

