/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* squarefree computation, Exercise 14.27 page 415,
   Gathen, Modern Computer Algebra, 2nd edition */

/**
> `struct ll_vec_lili *ll_upol_squarefree_naive(R *a,uma na);`
  Return a linked list containing squarefree factors of `a`.
  The factors may not be irreducible.
  The naive algorithm is used.
*/

struct ll_vec_lili *ll_upol_squarefree_naive(R *a,uma na) {
  R *buf,*disc,*gcd;
  uma nbuf,nbuf2,ndisc,ndisc2,ngcd,ngcd2,nsep,tmp;
  struct ll_vec_lili *ret,*sep,*rad;

  if ( na == 1 || na == 2 ) { /* degree 1 or 2 polynomials */
    ret = ll_vec_lili_node_malloc(na);
    ll_vec_copy(ret->a,a,na);
    return ret;
  }
 
  /* compute the derivative of a(X) */
  ret = NULL;
  nbuf2 = na;
  buf = ll_vec_malloc_init(nbuf2);
  ll_upol_derivative(buf,a,na);
  for( nbuf = nbuf2 - 2 ; nbuf >= 0 && R_iszero(buf[nbuf]) ; nbuf-- );
  nbuf++;
  if ( nbuf <= 0 ) { /* a(X) is a p-th power */
    R *pthroot;
    uma npthroot;
    pthroot = buf;
    ll_upol_pthroot(pthroot,&npthroot,a,na);
    ret = ll_upol_squarefree(pthroot,npthroot);
    goto ll_upol_squarefree_free_buf;
  }
 
  /* compute gcd(a(X),a'(X)) */
  ndisc2 = na;
  disc = ll_vec_malloc_init(ndisc2);
  ll_upol_gcd(disc,&ndisc,a,na,buf,nbuf);
  if ( ndisc == 1 ) { /* disc = 1 => a(X) is squarefree */
    ret = ll_vec_lili_node_malloc(na);
    ll_vec_copy(ret->a,a,na);
    goto ll_upol_squarefree_free_disc;
  }

  /* get the separable part */
  nsep = na - ndisc + 1;
  sep = ll_vec_lili_node_malloc(nsep);
  ll_upol_quo(sep->a,a,na,disc,ndisc);
  ret = sep;

  /* is there an inseparable part? */
  ll_upol_pow_mod_uma(buf,&nbuf,sep->a,nsep,na - 1,disc,ndisc);
  ngcd2 = MIN(ndisc,nbuf);
  gcd = ll_vec_malloc_init(ngcd2);
  ll_upol_gcd(gcd,&ngcd,disc,ndisc,buf,nbuf);
  nbuf = ndisc - ngcd + 1;
  ll_upol_quo(buf,disc,ndisc,gcd,ngcd);
  ll_upol_pthroot(buf,&tmp,buf,nbuf);
  rad = ll_upol_squarefree(buf,tmp);
  ll_vec_lili_node_attach_before(rad,sep);
  ll_vec_free(gcd,ngcd2);

ll_upol_squarefree_free_disc:
  ll_vec_free(disc,ndisc2);

ll_upol_squarefree_free_buf:
  ll_vec_free(buf,nbuf2);

  return ret;
}

/**
> `struct ll_vec_lili *ll_upol_squarefree(R *a,uma na);`
  Return a linked list containing squarefree factors of `a`.
  The factors may not be irreducible.
*/

struct ll_vec_lili *ll_upol_squarefree(R *a,uma na) {
#ifdef MY_UPOL_SQUAREFREE
  return MY_UPOL_SQUAREFREE(a,na);
#else
  return ll_upol_squarefree_naive(a,na);
#endif
}

