/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* equal-degree splitting, Algorithm 14.27 page 415,
   Gathen, Modern Computer Algebra, 2nd edition */

/**
> `void ll_upol_trace_poly(R *r,uma *nr,uma d,R *a,uma na,\
R *m,uma nm)`
  Set `r` of size `nr` to be the trace polynomial of degree `d` of
  `a` of size `na` in the sense of exercise 14.16, page 413 of
  `Modern Computer Algebra, 2nd edition` by Von zur Gathen and
  Gerhard.
*/

void ll_upol_trace_poly(R *r,uma *nr,R *a,uma na,uma d,R *m,uma nm) {
#if defined(RING_FF) && defined(RING_CHAR2)
  uma i,max;
  R *acc,*f,*g;
  uma nacc,nacc2,nf,nf2,ng,ng2;

  nacc2 = MAX(na,nm);
  acc = ll_vec_malloc_init(nacc2);
  nf2 = crit_mul(2,nm);
  f = ll_vec_malloc_init(nf2);
  ng2 = nf2;
  g = ll_vec_malloc_init(ng2);
  
  ll_upol_rem(acc,&nacc,a,na,m,nm);
  ll_vec_copy(f,acc,nacc);
  nf = nacc;
  max = crit_mul(R_ext_degree,d);
  for( i = 1 ; i < max ; i++ ) {
    ll_upol_square(g,f,nf);
    ng = 2 * nf - 1;
    ll_upol_rem(f,&nf,g,ng,m,nm);
    ll_array_add(acc,acc,nacc,f,nf);
    nacc = ll_vec_true_size(acc,MAX(nacc,nf));
  }
  nr[0] = nacc;
  ll_vec_copy(r,acc,nacc);
  ll_vec_free(acc,nacc2);
  ll_vec_free(f,nf2);
  ll_vec_free(g,ng2);
#else
  (void)r;
  (void)nr;
  (void)d;
  (void)a;
  (void)na;
  (void)m;
  (void)nm;
#endif
}

/**
> `void ll_upol_cantor_zassenhaus(R *r,R *a,uma na,uma d);`
  Set `r` to be a proper factor of `a`.
  All irreducible factors of `a` must have degree `d`.
  Note that `r` may not be irreducible and that `a` must
  be monic and has degree `na - 1`.
  The algorithm of Cantor and Zassenhaus is used.
*/

void ll_upol_cantor_zassenhaus(R *r,uma *nr,R *a,uma na,uma d) {
#ifdef RING_FF
  R *f,*g;
  uma nf,nf2,ng,ng2,ngcd;
#ifndef RING_CHAR2 
  mpz_t expo;
#endif

#ifndef RING_CHAR2 
  mpz_init(expo);
  mpz_set(expo,R_card);
  mpz_pow_ui(expo,expo,__CAST(unsigned long,d));
  mpz_sub_ui(expo,expo,1);
  mpz_divexact_ui(expo,expo,2);
#endif
  nf2 = na;
  f = ll_vec_malloc_init(nf2);
  ng2 = na;
  g = ll_vec_malloc_init(ng2);
  /* FIXME: make this loop finite for sure */
  for(;;) {
    ll_vec_random(f,nf2 - 1);
    nf = ll_vec_true_size(f,nf2 - 1);
    if ( nf <= 1 ) continue;
    ll_upol_gcd(r,&ngcd,a,na,f,nf);
    if ( ll_vec_true_size(r,ngcd) > 1 ) break;
#ifndef RING_CHAR2
    ll_upol_pow_mod(g,&ng,f,nf,expo,a,na);
#else
    ll_upol_trace_poly(g,&ng,f,nf,d,a,na);
#endif
    if ( ng <= 1 ) continue;
    R_sub(g[0],g[0],R_one);
    if ( ng <= 1 ) continue;
    ll_upol_gcd(r,&ngcd,g,ng,a,na); 
    if ( ngcd > 1 && ngcd < na ) break;
  }
  ll_upol_normalize(r,ngcd);
  nr[0] = ngcd;
  ll_vec_free(f,nf2);
  ll_vec_free(g,ng2);
#ifndef RING_CHAR2
  mpz_clear(expo);
#endif
#else
  (void)r;
  (void)nr;
  (void)a;
  (void)na;
  (void)d;
#endif
}

/**
> `void ll_upol_eqdegsplit(R *r,R *a,uma na);`
  Set `r` to be a proper factor of `a`.
  Note that `r` may not be irreducible.
  Note that `a` *must* be monic and have degree `na - 1`.
*/

void ll_upol_eqdegsplit(R *r,uma *nr,R *a,uma na,uma d) {
#ifdef MY_UPOL_EQDEGSPLIT
  MY_UPOL_EQDEGSPLIT(r,nr,a,na,d);
#else
  ll_upol_cantor_zassenhaus(r,nr,a,na,d);
#endif
}

