/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `void ll_upol_quorem_naive(R *r,uma *nr,R *q,R *a,uma na,R *b,\
  uma nb);`
  Set `r` of size `nr` and `q` to be the rest and the quotient of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1`.
  The schoolbook algorithm is used.
  You can have `r == a`.
*/

void ll_upol_quorem_naive(R *r,uma *nr,R *q,R *a,uma na,R *b,uma nb) {
  R inv;
  uma dq,dq2,dr,i,j;
  
  R_init(inv);
  if ( R_inv(inv,b[nb - 1]) ) {
    ll_die("leading coefficient not invertible\n"
           "cannot do Euclidean division");
  }
  dq = na - nb;
  dr = na - 1;
  if ( r != a ) ll_vec_copy(r,a,na);
  for( ; dr >= nb - 1 ; ) {
    R_mul(q[dq],r[dr],inv);
    R_zero(r[dr]);
    dr--;
    for( dq2 = dq - 1,i = dr,j = nb - 2 ; j >= 0 ; i--,j-- ) {
      R_sub_mul(r[i],b[j],q[dq]);
      if ( dq2 >= 0 && R_iszero(r[i]) ) {
        R_zero(q[dq2]);
        dq2--;
        dr--;
      }
      else {
        for( i--,j-- ; j >= 0 ; i--,j-- ) {
          R_sub_mul(r[i],b[j],q[dq]);
        }
        break;
      }
    }
    dq = dq2;
  }
  for( ; dr >=0 && R_iszero(r[dr]) ; dr-- );
  if ( dr < 0 ) {
    nr[0] = 1;
    R_zero(r[0]);
  }
  else {
    nr[0] = dr + 1;
  }
  R_clear(inv);
}

/**
> `void ll_upol_quo_naive(R *q,R *a,uma na,R *b,uma nb);`
  Set `q` to be the quotient of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1`.
  The schoolbook algorithm is used.
*/

void ll_upol_quo_naive(R *q,R *a,uma na,R *b,uma nb) {
  R inv,*r;
  uma dq2,dq,dr,i,j;

  r = ll_vec_malloc_init(na);
  R_init(inv);
  if ( R_inv(inv,b[nb - 1]) ) {
    ll_die("leading coefficient not invertible\n"
           "cannot do Euclidean division");
  }
  dq = na - nb;
  dr = na - 1;
  ll_vec_copy(r,a,na);
  for( ; dr >= nb - 1 ; ) {
    R_mul(q[dq],r[dr],inv);
    R_zero(r[dr]);
    dr--;
    for( dq2 = dq - 1,i = dr,j = nb - 2 ; j >= 0 ; i--,j-- ) {
      R_sub_mul(r[i],b[j],q[dq]);
      if ( dq2 >= 0 && R_iszero(r[i]) ) {
        R_zero(q[dq2]);
        dq2--;
        dr--;
      }
      else {
        for( i--,j-- ; j >= 0 ; i--,j-- ) {
          R_sub_mul(r[i],b[j],q[dq]);
        }
        break;
      }
    }
    dq = dq2;
  }
  R_clear(inv);
  ll_vec_free(r,na);
}

/**
> `void ll_upol_rem_naive(R *r,uma *nr,R *a,uma na,R *b,uma nb);`
  Set `r` of size `nr` to be the rest of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1`.
  The schoolbook algorithm is used.
  You can have `r == a`.
*/

void ll_upol_rem_naive(R *r,uma *nr,R *a,uma na,R *b,uma nb) {
  R inv,q;
  uma dr,i,j;
  
  R_init(inv);
  R_init(q);
  if ( R_inv(inv,b[nb - 1]) ) {
    ll_die("leading coefficient not invertible\n"
           "cannot do Euclidean division");
  }
  dr = na - 1;
  if ( r != a ) ll_vec_copy(r,a,na);
  for( ; dr >= nb - 1 ; ) {
    R_mul(q,r[dr],inv);
    R_zero(r[dr]);
    dr--;
    for( j = nb - 2 ; j >= 0 ; j--,dr-- ) {
      R_sub_mul(r[dr],b[j],q);
      if ( !R_iszero(r[dr]) ) break;
    }
    for( i = dr - 1,j-- ; j >= 0 ; i--,j-- ) {
      R_sub_mul(r[i],b[j],q);
    }
  }
  for( ; dr >= 0 && R_iszero(r[dr]) ; dr-- );
  if ( dr < 0 ) {
    nr[0] = 1;
    R_zero(r[0]);
  }
  else {
    nr[0] = dr + 1;
  }
  R_clear(q);
  R_clear(inv);
}

/* TODO: implement fast euclidean division: Moenck, Borodin */

/**
> `void ll_upol_quorem(R *r,uma *nr,R *q,R *a,uma na,R *b,uma nb);`
  Set `r` of size `nr` and `q` to be the rest and the quotient of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1` and that `r`
  must point to a region of at least `na` allocated `R`'s.
*/

void ll_upol_quorem(R *r,uma *nr,R *q,R *a,uma na,R *b,uma nb) {
#ifdef MY_UPOL_QUOREM
  MY_UPOL_QUOREM(r,nr,q,a,na,b,nb);
#else
  ll_upol_quorem_naive(r,nr,q,a,na,b,nb);
#endif
}

/**
> `void ll_upol_quo(R *q,R *a,uma na,R *b,uma nb);`
  Set `q` to be the quotient of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1` and that `r`
  must point to a region of at least `na` allocated `R`'s.
*/

void ll_upol_quo(R *q,R *a,uma na,R *b,uma nb) {
#ifdef MY_UPOL_QUO
  MY_UPOL_QUO(q,a,na,b,nb);
#else
  ll_upol_quo_naive(q,a,na,b,nb);
#endif
}

/**
> `void ll_upol_rem(R *r,uma *nr,R *a,uma na,R *b,uma nb);`
  Set `r` of size `nr` to be the rest of
  the Euclidean division of `a` of size `na` by `b` of size `nb`.
  Note that the degree of `b` *must* be `nb - 1` and that `r`
  must point to a region of at least `na` allocated `R`'s.
*/

void ll_upol_rem(R *r,uma *nr,R *a,uma na,R *b,uma nb) {
#ifdef MY_UPOL_REM
  MY_UPOL_REM(r,nr,a,na,b,nb);
#else
  ll_upol_rem_naive(r,nr,a,na,b,nb);
#endif
}

