/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* Partial Extended Euclidean Algorithm, Modern Computer Algebra,
   Algorithm 3.6, page 46 */

/**
> `void ll_upol_partial_eea_naive(R *r,uma *nr,R *u,uma *nu,\
R *v,uma *nv,R *a,uma na,R *b,uma nb,uma k);`
  Set `r` of size `nr`, `u` of size `nu` and `v` of size `nv`
  to be the polynomials involved in the
  Bezout's identity produced by the extended Euclidean
  algorithm when the remainder has degree at most `k` with input
  `a` of size `na` and `b` of size `nb`. Namely it returns
  `r`, `u` and `v` such that `u(X) a(X) + v(X) b(X) = r(X)`
  such that $r$ is the first remainder which has degree at most `k`.
  Note that `u` must point to a region of at least `nb` allocated
  `R`'s and `v` must point to a region of at least `na` allocated
  `R`'s.
  The naive partial extended Euclidean algorithm is used.
*/

void ll_upol_partial_eea_naive(R *r,uma *nr,
                               R *u,uma *nu,R *v,uma *nv,
                               R *a,uma na,R *b,uma nb,uma k)
{
  R *f,*g,*q,*xchg,*s,*ss,*t,*tt,*tmp;
  uma nf0,nf,ng0,ng,nq,nq0,nxchg,ntmp0,ntmp;
  uma ns0,nss0,ns,nss,nt0,ntt0,nt,ntt;

  /*if ( na >= nb ) {*/
    f = ll_vec_malloc_copy(a,na); nf = na;
    g = ll_vec_malloc_copy(b,nb); ng = nb;
  /*
  }
  else {
    g = ll_vec_malloc_copy(a,na); ng = na;
    f = ll_vec_malloc_copy(b,nb); nf = nb;
  }*/

  /* deg f >= deg g
     cte = s f + t g */
  nf0 = nf;
  ng0 = ng;
  
  ns0 = ng0;
  s = ll_vec_malloc_init(ns0);
  nss0 = ng0;
  ss = ll_vec_malloc_init(nss0);
  R_castint(s[0],1);
  ns = 1;
  nss = 1;
  
  nt0 = nf0;
  t = ll_vec_malloc_init(nt0);
  ntt0 = nf0;
  tt = ll_vec_malloc_init(ntt0);
  R_castint(tt[0],1);
  nt = 1;
  ntt = 1;
 
  nq0 = nf0;
  q = ll_vec_malloc_init(nq0);

  ntmp0 = nf0;
  tmp = ll_vec_malloc_init(ntmp0);

  for( ; !ll_upol_iszero(g,ng) && nf > k ; ) {
    
    /* q <- quo(f,g) and f = rem(f,g) */
    nq = nf - ng + 1;
    ll_upol_quorem(f,&nf,q,f,nf,g,ng);
   
    /* s_{i + 1} <- s_{i - 1} - q s_i */
    ntmp0 = nss + nq - 1;
    ll_upol_mul(tmp,ss,nss,q,nq);
    ntmp = MIN(ntmp0,ng0);
    ll_array_sub(s,s,ns,tmp,ntmp);
    ns = ll_vec_true_size(s,MAX(ns,ntmp));
    XCHG(s,ss,xchg);
    XCHG(ns,nss,nxchg);
    
    /* t_{i + 1} <- t_{i - 1} - q t_i */
    ntmp0 = ntt + nq - 1;
    ll_upol_mul(tmp,tt,ntt,q,nq);
    ntmp = MIN(ntmp0,nf0);
    ll_array_sub(t,t,nt,tmp,ntmp);
    nt = ll_vec_true_size(t,MAX(nt,ntmp));
    XCHG(t,tt,xchg);
    XCHG(nt,ntt,nxchg);

    /* swap(f,g) */
    XCHG(f,g,xchg);
    XCHG(nf,ng,nxchg);
  }

  ll_vec_copy(u,s,ns);
  nu[0] = ns;
  ll_vec_copy(v,t,nt);
  nv[0] = nt;
  ll_vec_copy(r,f,nf);
  nr[0] = nf;

  ll_vec_free(s,ns0);
  ll_vec_free(ss,nss0);
  ll_vec_free(t,nt0);
  ll_vec_free(tt,ntt0);
  ll_vec_free(q,nq0);
  ll_vec_free(f,nf0);
  ll_vec_free(g,ng0);
  ll_vec_free(tmp,ntmp0);
}

/**
> `void ll_upol_partial_eea(R *r,uma *nr,R *u,uma *nu,\
R *v,uma *nv,R *a,uma na,R *b,uma nb);`
  Set `r` of size `nr`, `u` of size `nu` and `v` of size `nv`
  to be the polynomials involved in the
  Bezout's identity produced by the extended Euclidean
  algorithm when the remainder has degree at most `k` with input
  `a` of size `na` and `b` of size `nb`. Namely it returns
  `r`, `u` and `v` such that `u(X) a(X) + v(X) b(X) = r(X)`
  such that $r$ is the first remainder which has degree at most `k`.
  Note that `u` must point to a region of at least `nb` allocated
  `R`'s and `v` must point to a region of at least `na` allocated
  `R`'s.
*/

void ll_upol_partial_eea(R *r,uma *nr,R *u,uma *nu,R *v,uma *nv,
                         R *a,uma na,R *b,uma nb,uma k)
{
#ifdef MY_UPOL_PARTIAL_EEA
  MY_UPOL_PARTIAL_EEA(r,nr,u,nu,v,nv,a,na,b,nb,k);
#else
  ll_upol_partial_eea_naive(r,nr,u,nu,v,nv,a,na,b,nb,k);
#endif
}

