/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
==OVERVIEW==

It is easy to control what algorithms are used by `decoding`, for
example if you want to replace unvariate polynomial multiplication
with your own function or if the ring you are implementing has
caracteristic 2.

==DESCRIPTION==

===TYPE AND NAME OF A RING===

The two macros defined below control the name of the ring and
the `C` type used to represent its elements. These macros are
at the disposal of the end-user.
*/

/**
> `#define R Name`
  `R` contains the name `Name` of the ring you are implementing. See
  `decoding-rings`(3) for details.
*/

#ifdef R
#undef R
#endif

/**
> `#define RING_NAME Name`
  `RING_NAME` contains the name you want to give the ring you are
  using. It must a valid C type identifier as it will be used by
  you and `decoding` to designate the C type corresponding the
  the ring you want to use. See `decoding-alphabets`(3) for
  more details.
*/

#ifdef RING_NAME
#undef RING_NAME
#endif

/**
===MACROS THAT CONTROL OPTIMIZATIONS===

The file `include/decoding/rings/GF5.c` is an example of how to
use macros defined below. We first list all the macros concerning
the behavior of the ring. When these macros are defined certain
optimizations are done by `decoding`. These macros are at the
disposal of the library developpers but can be used by the end-user
if he knows what he is doing.
*/

/**
> `#define RING_COMM`
  Indicates that the ring is commutative.
*/

#ifdef RING_COMM
#undef RING_COMM
#endif

/**
> `#define RING_CHAR_UMA`
  Indicates that the characteristic of the ring can hold within
  a `uma` integer.
*/

#ifdef RING_CHAR_UMA
#undef RING_CHAR_UMA
#endif

/**
> `#define RING_CHAR_MPZ`
  Indicates that the characteristic of the ring cannot hold,
  *a priori* within a `uma` integer. If it is the case then
  `R_uma_char` will be `-1` otherwise `R_uma_char` *must* hold the
  characteristic of the ring *even* if `RING_CHAR_MPZ` defined.
*/

#ifdef RING_CHAR_MPZ
#undef RING_CHAR_MPZ
#endif

/**
> `#define RING_CHAR2`
  Indicates that the ring has characteristic 2.
*/

#ifdef RING_CHAR2
#undef RING_CHAR2
#endif

/**
> `#define RING_FF`
  Indicates that the ring is in fact a finite field.
*/

#ifdef RING_FF
#undef RING_FF
#endif

/**
> `#define RING_MODN_WORD`
  Tell `decoding` that the ring is `Z/nZ` where `n` holds
  within a machine word. It can corresponds, for example, to the
  `unsigned char`, `unsigned short`, `unsigned int` or
  `unsigned long` C type. The `WORD_TYPE` macro must
  be defined to the C type that represents the elements of
  your ring.
*/

#ifdef RING_MODN_WORD
#undef RING_MODN_WORD
#endif

/**
> `#define RING_CHAR2_WORD`
  Tell `decoding` that the ring is `F_{2^n}` where `n` holds
  within a machine word. It can corresponds, for example, to the
  `unsigned char`, `unsigned short`, `unsigned int` or
  `unsigned long` C type. The `WORD_TYPE` macro must
  be defined to the C type that represents the elements of
  your ring.
*/

#ifdef RING_CHAR2_WORD
#undef RING_CHAR2_WORD
#endif

#ifdef WORD_TYPE
#undef WORD_TYPE
#endif

/**
===MACROS FOR FUNCTION REPLACEMENT===

Each macro below can be used so that the user supplies his own
function in place of the one provided by `decoding`.
For example if the end-user wants to use his own univariate
polynomial multiplication algorithm, he can use the `MY_UPOL_MUL`
macro.

``
#include <decoding/rings/some_ring.c>

#define MY_UPOL_MUL my_super_multiplication
void my_super_multiplication(R *r,R *a,uma na,R *b,uma nb) {
  \/\* my super multiplication function *\/
}

#include <decoding/algos.c>

int main(void) {
  \/\* some stuff *\/
  return 0;
}
``

All the following macros are to be used as follows.

``
#define MACRO myfunc
``

In the above example `MACRO` is a macro listed below and
myfunc is the name of the end-user's function which is to be used.
See `decoding-lowlevels`(3) for the prototype that must have `myfunc`.
The following macros are at the disposal of the end-user.
*/

/**
- `MY_UPOL_EVAL` for replacing `ll_upol_eval`.
*/

#ifdef MY_UPOL_EVAL
#undef MY_UPOL_EVAL
#endif

/**
- `MY_UPOL_MUL` for replacing `ll_upol_mul`.
*/

#ifdef MY_UPOL_MUL
#undef MY_UPOL_MUL
#endif

/**
- `MY_UPOL_SQUARE` for replacing `ll_upol_square`.
*/

#ifdef MY_UPOL_SQUARE
#undef MY_UPOL_SQUARE
#endif

/**
- `MY_UPOL_PTHROOT` for replacing `ll_upol_pthroot`.
*/

#ifdef MY_UPOL_PTHROOT
#undef MY_UPOL_PTHROOT
#endif

/**
- `MY_UPOL_SHIFT` for replacing `ll_upol_shift`.
*/

#ifdef MY_UPOL_SHIFT
#undef MY_UPOL_SHIFT
#endif

/**
- `MY_UPOL_QUOREM` for replacing `ll_upol_quorem`.
*/

#ifdef MY_UPOL_QUOREM
#undef MY_UPOL_QUOREM
#endif

/**
- `MY_UPOL_QUO` for replacing `ll_upol_quo`.
*/

#ifdef MY_UPOL_QUO
#undef MY_UPOL_QUO
#endif

/**
- `MY_UPOL_REM` for replacing `ll_upol_rem`.
*/

#ifdef MY_UPOL_REM
#undef MY_UPOL_REM
#endif

/**
- `MY_UPOL_GCD` for replacing `ll_upol_gcd`.
*/

#ifdef MY_UPOL_GCD
#undef MY_UPOL_GCD
#endif

/**
- `MY_UPOL_EEA` for replacing `ll_upol_eea`.
*/

#ifdef MY_UPOL_EEA
#undef MY_UPOL_EEA
#endif

/**
- `MY_UPOL_PARTIAL_EEA` for replacing `ll_upol_partial_eea`.
*/

#ifdef MY_UPOL_PARTIAL_EEA
#undef MY_UPOL_PARTIAL_EEA
#endif

/**
- `MY_UPOL_INTERPOLATION` for replacing `ll_upol_interpolation`.
*/

#ifdef MY_UPOL_INTERPOLATION
#undef MY_UPOL_INTERPOLATION
#endif

/**
- `MY_UPOL_ROOTS_FQ` for replacing `ll_upol_roots_fq`.
*/

#ifdef MY_UPL_ROOTS_FQ
#undef MY_UPL_ROOTS_FQ
#endif

/**
- `MY_UPOL_ROOTS` for replacing `ll_upol_roots`.
*/

#ifdef MY_UPOL_ROOTS
#undef MY_UPOL_ROOTS
#endif

/**
- `MY_UPOL_POW_MOD` for replacing `ll_upol_pow_mod`.
*/

#ifdef MY_UPOL_POW_MOD
#undef MY_UPOL_POW_MOD
#endif

/**
- `MY_UPOL_POW_MOD_UMA` for replacing `ll_upol_pow_mod_uma`.
*/

#ifdef MY_UPOL_POW_MOD_UMA
#undef MY_UPOL_POW_MOD_UMA
#endif

/**
- `MY_UPOL_DERIVATIVE` for replacing `ll_upol_derivative`.
*/

#ifdef MY_UPOL_DERIVATIVE
#undef MY_UPOL_DERIVATIVE
#endif

/**
- `MY_UPOL_DIV_BY_AFFINE` for replacing `ll_upol_div_by_affine`.
*/

#ifdef MY_UPOL_DIV_BY_AFFINE
#undef MY_UPOL_DIV_BY_AFFINE
#endif

/**
- `MY_DBPOL_MUL` for replacing `ll_dbpol_mul`.
*/

#ifdef MY_DBPOL_MUL
#undef MY_DBPOL_MUL
#endif

/**
- `MY_DBPOL_SHIFT` for replacing `ll_dbpol_shift`.
*/

#ifdef MY_DBPOL_SHIFT
#undef MY_DBPOL_SHIFT
#endif

/**
- `MY_DBPOL_XROOTS` for replacing `ll_dbpol_Xroots`.
*/

#ifdef MY_DBPOL_XROOTS
#undef MY_DBPOL_XROOTS
#endif

/**
- `MY_VEC_ADD` for replacing `ll_vec_add`.
*/

#ifdef MY_VEC_ADD
#undef MY_VEC_ADD
#endif

/**
- `MY_VEC_NEG` for replacing `ll_vec_neg`.
*/

#ifdef MY_VEC_NEG
#undef MY_VEC_NEG
#endif

/**
- `MY_VEC_SUB` for replacing `ll_vec_sub`.
*/

#ifdef MY_VEC_SUB
#undef MY_VEC_SUB
#endif

/**
- `MY_VEC_SUB_SUB` for replacing `ll_vec_sub_sub`.
*/

#ifdef MY_VEC_SUB_SUB
#undef MY_VEC_SUB_SUB
#endif

/**
- `MY_VEC_ADD_MUL` for replacing `ll_vec_add_mul`.
*/

#ifdef MY_VEC_ADD_MUL
#undef MY_VEC_ADD_MUL
#endif

/**
- `MY_VEC_MUL_BY_CTE` for replacing `ll_vec_mul_by_cte`.
*/

#ifdef MY_VEC_MUL_BY_CTE
#undef MY_VEC_MUL_BY_CTE
#endif

/**
- `MY_VEC_ADD_MUL_BY_CTE` for replacing `ll_vec_add_mul_by_cte`.
*/

#ifdef MY_VEC_ADD_MUL_BY_CTE
#undef MY_VEC_ADD_MUL_BY_CTE
#endif

/**
- `MY_VEC_SUB_MUL` for replacing `ll_vec_sub_mul`.
*/

#ifdef MY_VEC_SUB_MUL
#undef MY_VEC_SUB_MUL
#endif

/**
- `MY_VEC_PROD` for replacing `ll_vec_prod`.
*/

#ifdef MY_VEC_PROD
#undef MY_VEC_PROD
#endif

/**
===MEMORY MANAGEMENT===

`Decoding` provides no garbage collectors by default.
The memory is manually allocated and freed with the
standard libc functions `stdlib.h`(0P). You can replace
the memory management functions with your own or
use an external garbage collector if you wish with the
`MY_*ALLOC` macros.

> `#define MY_MALLOC myfunc`
  Replace the `malloc` function by `myfunc`. By default
  the libc `malloc` is used. `myfunc` must have the same
  prototype as `malloc`.
  See `malloc`(3) for more details.

> `#define MY_REALLOC myfunc`
  Replace the `realloc` function by `myfunc`. By default
  the libc `realloc` is used. `myfunc` must have the same
  prototype as `realloc`.
  See `realloc`(3) for more details.

> `#define MY_CALLOC myfunc`
  Replace the `calloc` function by `myfunc`. By default
  the libc `calloc` is used. `myfunc` must have the same
  prototype as `calloc`.
  See `calloc`(3) for more details.

> `#define MY_FREE myfunc`
  Replace the `free` function by `myfunc`. By default
  the libc `free` is used. `myfunc` must have the same
  prototype as `free`.
  See `free`(3) for more details.

==AUTHOR==

Written by Guillaume Quintin (coincoin169g@gmail.com).

==SEE ALSO==

`decoding-rings`(3),`decoding-lowlevels`(3),
MAGMA(http://magma.maths.usyd.edu.au/magma/),
`malloc`(3), `realloc`(3), `calloc`(3), `free`(3),
`stdlib.h`(0P).
*/

/* Other stuff to undef documented elsewhere */

#ifdef PRIME
#undef PRIME
#endif

#ifdef FP_PRIME
#undef FP_PRIME
#endif

#ifdef EXT_DEGREE
#undef EXT_DEGREE
#endif

#ifdef WORD_TYPE
#undef WORD_TYPE
#endif

#ifndef __DECODING_KEEP_MACROS

/* macros for base ring */
#include <decoding/rings_undef_macros.h>

/* macros for low levels functions */
#include <decoding/lowlevels_undef_macros.h>

/* macros for sugar functions */
#include <decoding/sugar_undef_macros.h>

#endif /* __DECODING_KEEP_MACROS */

#ifdef __cplusplus
} /* We close namespace decoding */
#endif
