/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
===`gf2n_word.c`===

Implement all finite fields `GF(2^n)` with `2 <= n <= 16`.
Note that it is a poor implementation (prefer `MPFQ`).
It is not intended to be efficient and will most likely
never be. Please use the `MPFQ` wrappers `mpfq_gf2n_wrapper.c`.
If `MPFQ` is not present on your system then `gf2n_word.c`
will be used instead. You do *not* have to change your source
code as `decoding` will automatically switch from
`mpfq_gf2n_wrapper.c` to `gf2n_word.c`.

You can give the extension degree with the `EXT_DEGREE` macro.
It must be an integer in the range `[2,16]`.

``
#define EXT_DEGREE 8
#include <decoding/rings/gf2n_word.c>
#include <decoding/algos.c>
``

In the example above, the default name of the ring will be `gf2_8`.
*/

#include <decoding/ring_init.h>

/* standard stuff for rings */
#ifdef RING_NAME
#define R RING_NAME
#else
#define R __C(gf2_,EXT_DEGREE)
#endif
#define RING_CHAR_UMA
#define RING_COMM
#define RING_CHAR2
#define RING_FF
#define RING_CHAR2_WORD

/* C genuine type choice, default to unsigned int */
#ifndef WORD_TYPE
#define WORD_TYPE unsigned int
#endif

#include <decoding/rings/gf2n_word_irr.h>

typedef WORD_TYPE *R_ptr;
typedef WORD_TYPE R[1];
typedef WORD_TYPE R_content;
const char *R_magma_decl = __S(R) " := GF(2," __S(EXT_DEGREE) ")";
mpz_t R_char;
const uma R_uma_char = 2;
mpz_t R_card;
const WORD_TYPE R_irr_poly = IRR_POLY(EXT_DEGREE);
const WORD_TYPE R_mask = __CAST(WORD_TYPE,-1) >>
                         (__SZ(WORD_TYPE) - EXT_DEGREE);
const uma R_ext_degree = EXT_DEGREE;
WORD_TYPE R_one[1] = { __CAST(WORD_TYPE,1) };

/*========================================================*/
/* The stuff is here                                      */
/*========================================================*/
#define R_reduce __C(R,_reduce)
#define R_internal_mul __C(R,_internal_mul)
#define R_internal_deg __C(R,_internal_deg)
#define R_internal_inv __C(R,_internal_inv)

WORD_TYPE R_reduce(WORD_TYPE a) {
  int i;
  for( i = 2 * EXT_DEGREE - 2 ; i >= EXT_DEGREE ; i-- ) {
    a ^= (R_irr_poly & (-((a >> i) & 1))) << (i - EXT_DEGREE);
  }
  return a;
}

WORD_TYPE R_internal_mul(WORD_TYPE a,WORD_TYPE b) {
  WORD_TYPE r;
  int i;
  r = 0;
  for( i = 0 ; i < EXT_DEGREE ; i++ ) {
    r ^= (a & (-((b >> i) & 1))) << i;
  }
  return R_reduce(r);
}

int R_internal_deg(WORD_TYPE a) {
  int d;
  d = -1;
  for( ; a ; d++,a >>= 1 );
  return d;
}

WORD_TYPE R_internal_inv(WORD_TYPE a) {
  int dr0, dr1, dr2;
  WORD_TYPE r0, r1, r2, v0, v1, v2, qv1, m;
  r0 = R_irr_poly;
  dr0 = EXT_DEGREE;
  r1 = a;
  dr1 = R_internal_deg(a);
  v0 = 0;
  v1 = 1;
  for( r2 = 1 ; r2 ; ) {
    r2 = r0;
    dr2= dr0;
    qv1 = 0;
    m = 1 << dr2;
    while( dr2 >= dr1 ) {
      r2 ^= r1 << (dr2 - dr1);
      qv1 ^= v1 << (dr2 - dr1);
      for( m >>= 1,dr2-- ; !(m & r2) && dr2 >= 0 ; dr2--,m >>= 1 );
    }
    v2= v0 ^ qv1;
    v0= v1;
    v1= v2;
    r0= r1;
    r1= r2;
    dr0= dr1;
    dr1= dr2;
  }
  return v0;
}
/*========================================================*/



void R_ring_init(void) {
  mpz_init(R_card);
  mpz_ui_pow_ui(R_card,2,EXT_DEGREE);
  mpz_init(R_char);
  mpz_set_si(R_char,2);
}

void R_ring_clear(void) {
  mpz_clear(R_card);
  mpz_clear(R_char);
}

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0;
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0;
  for(;;) {
    WORD_TYPE *b;
    b = va_arg(ap,WORD_TYPE*);
    if ( !b ) return;
    b[0] = 0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_swap(R r,R a) {
  R_content tmp;
  tmp = r[0];
  r[0] = a[0];
  a[0] = tmp;
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

void R_print(FILE *out,R a) {
  int i;
  if ( a[0] == 0 ) {
    fputc('0',out);
    return;
  }
  i = R_internal_deg(a[0]);
  for( ; i >= 0 ; i-- ) {
    fputc('0' + ((a[0] >> i) & 1),out);
  }
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0);
}

void R_zero(R r) {
  r[0] = 0;
}

void R_add(R r,R a,R b) {
  r[0] = a[0] ^ b[0];
}

void R_sub(R r,R a,R b) {
  r[0] = a[0] ^ b[0];
}

void R_sub_sub(R r,R a,R b) {
  r[0] = r[0] ^ a[0] ^ b[0];
}

void R_add_mul(R r,R a,R b) {
  r[0] = r[0] ^ R_internal_mul(a[0],b[0]);
}

void R_sub_mul(R r,R a,R b) {
  r[0] = r[0] ^ R_internal_mul(a[0],b[0]);
}

void R_neg(R r,R a) {
  r[0] = a[0];
}

void R_mul(R r,R a,R b) {
  r[0] = R_internal_mul(a[0],b[0]);
}

void R_square(R r,R a) {
  r[0] = R_internal_mul(a[0],a[0]);
}

void R_pthroot(R r,R a) {
  r[0] = a[0];
}

void R_castint(R r,int a) {
  if ( a >= 0 ) r[0] = a & 1; else r[0] = (-a) & 1;
}

void R_castuma(R r,uma a) {
  if ( a >= 0 ) r[0] = a & 1; else r[0] = (-a) & 1;
}

void R_random(R r) {
  r[0] = ll_random() & R_mask;
}

void R_cast_integer(R r,mpz_t a) {
  r[0] = mpz_fdiv_ui(a,2);
}

int R_inv(R r,R a) {
  if ( R_iszero(a) ) return -1;
  r[0] = R_internal_inv(a[0]);
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_next(R r,R a) {
  r[0] = (a[0] + 1) & R_mask;
}

#include <decoding/lowlevels_decl.h>

#undef R_reduce
#undef R_internal_mul
#undef R_internal_deg
#undef R_internal_inv

