/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
===`gfp_float.c`===

Implement all prime fields (`Z/pZ`) whose characteristic holds
within the significand of a `float`. Be careful, the size of a
`float` and the arithmetic is implementation-defined. *NO GUARANTY
IS GIVEN ON THE CORRETNESS OF THE ARITHMETIC* for `gfp_float.c`.
However it should work just fine on x86 and x86_64 processors.
Use `PRIME` to indicate the prime number defining the finite field.

``
#define PRIME 101
#include <decoding/rings/gfp_float.c>
#include <decoding/algos.c>
``

In the example above, the name of the finite field will be
`gf101`.

The prime number defining the field can also be accessed with
the `R_fp_char` variable of type `float`. Its numerical inverse is
stored in the `R_fp_char_inv` also of type `float`. For example
if `PRIME` is `101` we have
`gf101_fp_char == 101.0` and
`gf101_fp_char_inv == 0.00990098994`.
*/

#include <decoding/ring_init.h>

/* standard stuff for rings */
#ifdef RING_NAME
#define R RING_NAME
#else
#define R __C(gf,PRIME)
#endif
#define RING_MODN_WORD
#define RING_CHAR_UMA
#define RING_COMM
#define RING_FF
#define FP_PRIME __C(PRIME,.0f)

#define R_fp_char __C(R,_fp_char)
#define R_fp_char_inv __C(R,_fp_char_inv)

typedef float *R_ptr;
typedef float R[1];
typedef float R_content;
const char *R_magma_decl = __S(R) " := GF(" __S(PRIME) ")";
const uma R_uma_char = PRIME;
const uma R_ext_degree = 1;
mpz_t R_card;
mpz_t R_char;

#define R_fp_char __C(R,_fp_char)
#define R_fp_char_inv __C(R,_fp_char_inv)

const float R_fp_char = FP_PRIME;
const float R_fp_char_inv = 1.0 / FP_PRIME;
float R_one[1] = { __CAST(float,1) };

/*========================================================*/
/* Vectorization stuff: for now we let gcc vectorize      */
/*========================================================*/
#define R_reduce __C(R,_reduce)

float R_reduce(float a) {
  float q;
  q = __CAST(float,__CAST(int,a * R_fp_char_inv));
#if 0
  q = __CAST(float,__CAST(long,a / R_fp_char));
#endif
  return a - q * R_fp_char;
}
/*========================================================*/

void R_ring_init(void) {
  mpz_init(R_card);
  mpz_set_si(R_card,PRIME);
  mpz_init(R_char);
  mpz_set_si(R_char,PRIME);
}

void R_ring_clear(void) {
  mpz_clear(R_card);
  mpz_clear(R_char);
}

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0.0;
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0.0;
  for(;;) {
    float *b;
    b = va_arg(ap,float*);
    if ( !b ) return;
    b[0] = 0.0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_swap(R r,R a) {
  R_content tmp;
  tmp = r[0];
  r[0] = a[0];
  a[0] = tmp;
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

void R_print(FILE *out,R a) {
  fprintf(out,"%lu",__CAST(unsigned long,a[0]));
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0.0);
}

void R_zero(R r) {
  r[0] = 0.0;
}

void R_add(R r,R a,R b) {
  r[0] = R_reduce(a[0] + b[0]);
}

void R_sub(R r,R a,R b) {
  r[0] = R_reduce(a[0] + R_fp_char - b[0]);
}

void R_sub_sub(R r,R a,R b) {
  r[0] = R_reduce(r[0] + R_fp_char - a[0]);
  r[0] = R_reduce(r[0] + R_fp_char - b[0]);
}

void R_add_mul(R r,R a,R b) {
  r[0] = R_reduce(r[0] + R_reduce(a[0] * b[0]));
}

void R_sub_mul(R r,R a,R b) {
  r[0] = R_reduce(r[0] +
                  R_reduce((R_fp_char * R_fp_char) - (a[0] * b[0])));
}

void R_neg(R r,R a) {
  r[0] = R_reduce(R_fp_char - a[0]);
}

void R_mul(R r,R a,R b) {
  r[0] = R_reduce(a[0] * b[0]);
}

void R_square(R r,R a) {
  r[0] = R_reduce(a[0] * a[0]);
}

void R_pthroot(R r,R a) {
  r[0] = a[0];
}

void R_castint(R r,int a) {
  if ( a >= 0 ) {
    r[0] = __CAST(float,a % PRIME);
  }
  else {
    r[0] = __CAST(float,PRIME - ((-a) % PRIME));
  }
}

void R_castuma(R r,uma a) {
  if ( a >= 0 ) {
    r[0] = __CAST(float,a % PRIME);
  }
  else {
    r[0] = __CAST(float,PRIME - ((-a) % PRIME));
  }
}

void R_random(R r) {
  r[0] = __CAST(float,ll_random() % PRIME);
}

void R_cast_integer(R r,mpz_t a) {
  r[0] = mpz_fdiv_ui(a,PRIME);
}

int R_inv(R r,R a) {
  if ( a[0] == 0.0 ) return -1;
  r[0] = inv_mod(__CAST(unsigned long,a[0]),PRIME);
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_next(R r,R a) {
  r[0] = a[0] + 1.0;
  if ( r[0] == R_fp_char ) r[0] = 0.0;
}

#include <decoding/lowlevels_decl.h>

#undef R_reduce
#undef R_fp_char
#undef R_fp_char_inv

