/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
===`gfp_gmp.c`===

Implement all prime fields (`Z/pZ`).

``
#include <decoding/rings/gfp_gmp.c>
#include <decoding/algos.c>
``

In the example above, the name of the finite field will be `gfp_gmp`.

> `void R_ring_init(mpz_t prime);`
  Initialize the finite field as a prime field of char `p`.
  A call to this function is mandatory before doing anything else.

> `void R_ring_clear(void);`
  Free the memory occupied by the finite field.
  A call to this function is recommended when the ring is
  no more needed.
*/

#include <decoding/ring_init.h>

#ifdef RING_NAME
#define R RING_NAME
#else
#define R gfp_gmp
#endif

#define RING_CHAR_MPZ
#define RING_COMM
#define RING_FF
typedef mpz_ptr R_ptr;
typedef mpz_t R;
typedef mpz_t R_content;
const char *R_magma_decl = __S(R) " := GF(" __S(PRIME) ")";
uma R_uma_char;
const uma R_ext_degree = 1;
mpz_t R_card,R_char,R_one;

#define R_reduce_2p __C(R,_reduce_2p)
void R_reduce_2p(mpz_t r) {
  if ( mpz_cmp(r,R_char) >= 0 ) {
    mpz_sub(r,r,R_char);
  }
}

void R_ring_init(mpz_t prime) {
  mpz_init(R_char);
  mpz_set(R_char,prime);
  mpz_init(R_one);
  mpz_set_si(R_one,1);
  mpz_init(R_card);
  mpz_set(R_card,R_char);
  if ( mpz_cmp_si(R_char,uma_max) <= 0 ) {
    R_uma_char = __CAST(uma,mpz_get_ui(R_char));
  }
  else {
    R_uma_char = -1;
  }
}

void R_ring_clear(void) {
  mpz_clear(R_char);
  mpz_clear(R_one);
  mpz_clear(R_card);
}

void R_clear(R a) {
  mpz_clear(a);
}

void R_clears(R a,...) {
  va_list ap;
  va_start(ap,a);
  R_clear(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    R_clear(b);
  }
}

void R_init(R a) {
  mpz_init(a);
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);  
  R_init(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    mpz_init(b);
  }
}

void R_copy(R r,R a) {
  mpz_set(r,a);;
}

void R_swap(R r,R a) {
  mpz_swap(r,a);
}

void R_init_copy(R r,R a) {
  mpz_init_set(r,a);
}

void R_shallow_copy(R r,R a) {
  mpz_set(r,a);
}

void R_print(FILE *out,R a) {
  mpz_out_str(out,10,a);
}

int R_cmp(R a,R b) {
  return mpz_cmp(a,b);
}

int R_iszero(R a) {
  return !mpz_cmp_si(a,0);
}

void R_zero(R r) {
  mpz_set_si(r,0);
}

void R_add(R r,R a,R b) {
  mpz_add(r,a,b);
  R_reduce_2p(r);
}

void R_sub(R r,R a,R b) {
  mpz_add(r,a,R_char);
  mpz_sub(r,r,b);
  R_reduce_2p(r);
}

void R_sub_sub(R r,R a,R b) {
  mpz_add(r,r,R_char);
  mpz_sub(r,r,a);
  R_reduce_2p(r);
  mpz_add(r,r,R_char);
  mpz_sub(r,r,b);
  R_reduce_2p(r);
}

void R_add_mul(R r,R a,R b) {
  mpz_addmul(r,a,b);
  mpz_mod(r,r,R_char);
}

void R_sub_mul(R r,R a,R b) {
  mpz_submul(r,a,b);
  mpz_mod(r,r,R_char);
}

void R_neg(R r,R a) {
  mpz_sub(r,R_char,a);
  R_reduce_2p(r);
}

void R_mul(R r,R a,R b) {
  mpz_mul(r,a,b);
  mpz_mod(r,r,R_char);
}

void R_square(R r,R a) {
  mpz_mul(r,a,a);
  mpz_mod(r,r,R_char);
}

void R_pthroot(R r,R a) {
  mpz_set(r,a);
}

void R_castint(R r,int a) {
  mpz_set_si(r,a);
}

void R_castuma(R r,uma a) {
  mpz_set_si(r,a);
}

void R_random(R r) {
  mpz_urandomm(r,__gmp_rands,R_char);
}

void R_cast_integer(R r,mpz_t a) {
  mpz_mod(r,a,R_char);
}

int R_inv(R r,R a) {
  return !mpz_invert(r,a,R_char);
}

int R_div(R r,R a,R b) {
  R c;
  R_init(c);
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  R_clear(c);
  return 0;
}

void R_next(R r,R a) {
  mpz_add_ui(r,a,1);
  R_reduce_2p(r);
}

#include <decoding/lowlevels_decl.h>

#undef R_reduce_2p

