/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
===`gfp_word.c`===

Implement all prime fields (`Z/pZ`) whose char holds
within a machine word. Use `PRIME` to indicate the modulus of
the finite field you want to construct.

``
#define PRIME 101
#include <decoding/rings/gfp_word.c>
#include <decoding/algos.c>
``

In the example above, the name of the finite field will be `gf101`.
You can use the `WORD_TYPE` macro to specify which C type to use
to represent the elements of the finite field. The given type
*must* be an unsigned type. By default the `unsigned int` type is
used.

``
#define PRIME 101
#define WORD_TYPE unsigned short
#include <decoding/rings/gfp_word.c>
#include <decoding/algos.c>
``
*/

#include <decoding/ring_init.h>

/* standard stuff for rings */
#ifdef RING_NAME
#define R RING_NAME
#else
#define R __C(gf,PRIME)
#endif

/* C genuine type choice, default to unsigned int */
#ifndef WORD_TYPE
#define WORD_TYPE unsigned int
#endif

#define RING_MODN_WORD
#define RING_CHAR_UMA
#define RING_COMM
#define RING_FF

typedef WORD_TYPE *R_ptr;
typedef WORD_TYPE R[1];
typedef WORD_TYPE R_content;
const char *R_magma_decl = __S(R) " := GF(" __S(PRIME) ")";
WORD_TYPE R_one[1] = { __CAST(WORD_TYPE,1) };
const uma R_ext_degree = 1;
const uma R_uma_char = PRIME;
mpz_t R_card,R_char;

/*========================================================*/
/* Vectorization stuff: for now we let gcc vectorize      */
/*========================================================*/
#define R_reduce_2p __C(R,_reduce_2p)
#define R_reduce_p2 __C(R,_reduce_p2)
#define R_negate __C(R,_negate)
#define PRIME_BITS CEIL_LOG2(PRIME)
#define PRIME_INV ((1 << (2 * PRIME_BITS + 1)) / PRIME)

WORD_TYPE R_reduce_2p(WORD_TYPE a) {
#ifdef __SSE41
  WORD_TYPE tmp;
  tmp = a - PRIME;
  return MIN(a,tmp);
#else
  WORD_TYPE mask;
  mask = -(__CAST(WORD_TYPE,PRIME - 1 - a) >> (__SZ(WORD_TYPE) - 1));
  return a - (PRIME & mask);
#endif
#if 0
  return a % PRIME;
#endif
}

WORD_TYPE R_reduce_p2(WORD_TYPE a) {
  WORD_TYPE a1,b,c,prod;
  a1 = a >> (PRIME_BITS - 2);
  prod = a1 * PRIME_INV;
  b = prod >> (PRIME_BITS + 3);
  prod = b * PRIME;
  c = a - prod;
  return R_reduce_2p(c);
#if 0
  return a % PRIME;
#endif
}

WORD_TYPE R_negate(WORD_TYPE a) {
  return R_reduce_2p(PRIME - a);
}
/*========================================================*/



void R_ring_init(void) {
  mpz_init(R_char);
  mpz_set_si(R_char,PRIME);
  mpz_init(R_card);
  mpz_set_si(R_card,PRIME);
}

void R_ring_clear(void) {
  mpz_clear(R_char);
  mpz_clear(R_card);
}

void R_clear(R a) {
  (void)a;
}

void R_clears(R a,...) {
  (void)a;
}

void R_init(R a) {
  a[0] = 0;
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);
  
  a[0] = 0;
  for(;;) {
    WORD_TYPE *b;
    b = va_arg(ap,WORD_TYPE*);
    if ( !b ) return;
    b[0] = 0;
  }
}

void R_copy(R r,R a) {
  r[0] = a[0];
}

void R_swap(R r,R a) {
  R_content tmp;
  tmp = r[0];
  r[0] = a[0];
  a[0] = tmp;
}

void R_init_copy(R r,R a) {
  r[0] = a[0];
}

void R_shallow_copy(R r,R a) {
  r[0] = a[0];
}

void R_print(FILE *out,R a) {
  fprintf(out,"%lu",__CAST(unsigned long,a[0]));
}

int R_cmp(R a,R b) {
  return (a[0] != b[0]);
}

int R_iszero(R a) {
  return (a[0] == 0);
}

void R_zero(R r) {
  r[0] = 0;
}

void R_add(R r,R a,R b) {
  r[0] = R_reduce_2p(a[0] + b[0]);
}

void R_sub(R r,R a,R b) {
  r[0] = R_reduce_2p(a[0] + R_negate(b[0]));
}

void R_sub_sub(R r,R a,R b) {
  r[0] = R_reduce_2p(r[0] + R_negate(a[0]));
  r[0] = R_reduce_2p(r[0] + R_negate(b[0]));
}

void R_add_mul(R r,R a,R b) {
  r[0] = R_reduce_2p(r[0] + R_reduce_p2(a[0] * b[0]));
}

void R_sub_mul(R r,R a,R b) {
  WORD_TYPE tmp;
  tmp = R_reduce_p2(a[0] * b[0]);
  tmp = R_negate(tmp);
  r[0] = R_reduce_2p(r[0] + tmp);
}

void R_neg(R r,R a) {
  r[0] = R_negate(a[0]);
}

void R_mul(R r,R a,R b) {
  r[0] = R_reduce_p2(a[0] * b[0]);
}

void R_square(R r,R a) {
  r[0] = R_reduce_p2(a[0] * a[0]);
}

void R_pthroot(R r,R a) {
  r[0] = a[0];
}

void R_castint(R r,int a) {
  if ( a >= 0 ) {
    r[0] = a % PRIME;
  }
  else {
    r[0] = R_negate((-a) % PRIME);
  }
}

void R_castuma(R r,uma a) {
  if ( a >= 0 ) {
    r[0] = a % PRIME;
  }
  else {
    r[0] = R_negate((-a) % PRIME);
  }
}

void R_random(R r) {
  unsigned long tmp;
  tmp = ll_random() % __CAST(unsigned long,PRIME);
  r[0] = __CAST(WORD_TYPE,tmp);
}

void R_cast_integer(R r,mpz_t a) {
  r[0] = mpz_fdiv_ui(a,PRIME);
}

int R_inv(R r,R a) {
  if ( a[0] == 0 ) return -1;
  r[0] = inv_mod(a[0],PRIME);
  return 0;
}

int R_div(R r,R a,R b) {
  R c;
  if ( R_inv(c,b) < 0 ) return -1;
  R_mul(r,a,c);
  return 0;
}

void R_next(R r,R a) {
  r[0] = a[0] + 1;
  if ( r[0] == PRIME ) r[0] = 0;
}

#include <decoding/lowlevels_decl.h>

#undef R_reduce_2p
#undef R_reduce_p2
#undef R_negate
#undef PRIME_BITS
#undef PRIME_INV

