/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
===`rationals_gmp.c`===

Implement the field or rational numbers.

``
#include <decoding/rings/rationals_gmp.c>
#include <decoding/algos.c>
``

In the example above, the name of the finite field will be
`rationals_gmp`.
*/

#include <decoding/ring_init.h>

/* standard stuff for rings */
#ifdef RING_NAME
#define R RING_NAME
#else
#define R rationals_gmp
#endif

#define RING_CHAR_MPZ
#define RING_COMM

typedef mpq_ptr R_ptr;
typedef mpq_t R;
typedef mpq_t R_content;
typedef mpz_t R_char_type;
const char *R_magma_decl = __S(R) " := Rationals()";

const uma R_ext_degree = 1;
const uma R_uma_char = 0;
mpz_t R_char;
mpz_t R_card;
mpq_t R_one;

void R_ring_init(void) {
  mpz_init(R_char);
  mpq_init(R_one);
  mpq_set_si(R_one,1,1);
  mpz_init(R_card);
}

void R_ring_clear(void) {
  mpz_clear(R_char);
  mpq_clear(R_one);
  mpz_clear(R_card);
}

void R_clear(R a) {
  mpq_clear(a);
}

void R_clears(R a,...) {
  va_list ap;
  va_start(ap,a);
  R_clear(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    R_clear(b);
  }
}

void R_init(R a) {
  mpq_init(a);
}

void R_inits(R a,...) {
  va_list ap;
  va_start(ap,a);  
  R_init(a);
  for(;;) {
    R_ptr b;
    b = va_arg(ap,R_ptr);
    if ( !b ) return;
    R_init(b);
  }
}

void R_copy(R r,R a) {
  mpq_set(r,a);;
}

void R_swap(R r,R a) {
  mpq_swap(r,a);
}

void R_init_copy(R r,R a) {
  mpq_init(r);
  mpq_set(r,a);
}

void R_shallow_copy(R r,R a) {
  mpq_set(r,a);
}

void R_print(FILE *out,R a) {
  mpq_out_str(out,10,a);
}

int R_cmp(R a,R b) {
  return mpq_cmp(a,b);
}

int R_iszero(R a) {
  return !mpq_cmp_si(a,0,1);
}

void R_zero(R r) {
  mpq_set_si(r,0,1);
}

void R_add(R r,R a,R b) {
  mpq_add(r,a,b);
}

void R_sub(R r,R a,R b) {
  mpq_sub(r,a,b);
}

void R_sub_sub(R r,R a,R b) {
  mpq_sub(r,r,a);
  mpq_sub(r,r,b);
}

void R_add_mul(R r,R a,R b) {
  mpq_t tmp;
  mpq_init(tmp);
  mpq_mul(tmp,a,b);
  mpq_add(r,r,tmp);
  mpq_clear(tmp);
}

void R_sub_mul(R r,R a,R b) {
  mpq_t tmp;
  mpq_init(tmp);
  mpq_mul(tmp,a,b);
  mpq_sub(r,r,tmp);
  mpq_clear(tmp);
}

void R_neg(R r,R a) {
  mpq_neg(r,a);
}

void R_mul(R r,R a,R b) {
  mpq_mul(r,a,b);
}

void R_square(R r,R a) {
  mpq_mul(r,a,a);
}

void R_pthroot(R r,R a) {
  mpq_set(r,a);
}

void R_castint(R r,int a) {
  mpq_set_si(r,a,1);
}

void R_castuma(R r,uma a) {
  mpq_set_si(r,a,1);
}

void R_random(R r) {
  unsigned long num,den;
  num = ll_random();
  den = ll_random();
  mpq_set_si(r,num,den);
  if ( ll_random() > (1ul << (__SZ(unsigned long) - 1)) ) {
    mpq_neg(r,r);
  }
}

void R_cast_integer(R r,mpz_t a) {
  mpq_set_z(r,a);
}

int R_inv(R r,R a) {
  if ( R_iszero(a) ) return -1;
  mpq_inv(r,a);
  return 0;
}

int R_div(R r,R a,R b) {
  if ( R_iszero(b) ) return -1;
  mpq_div(r,a,b);
  return 0;
}

void R_next(R r,R a) {
  mpq_t one;
  mpq_set_si(one,1,1);
  mpq_add(r,a,one);
  mpq_clear(one);
}

#include <decoding/lowlevels_decl.h>
#undef R_random_state

