/*
 * This file is part of decoding
 * Copyright 2012-2013  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/**
> `void rs_code_init(rs_code rs,uma n,uma k);`
  Initialize `rs` of length `n` and dimension `k`.
  This function must be called before doing anything else with `rs`.
  The support of `rs` will be a zero vector.
*/

void rs_code_init(rs_code rs,uma n,uma k) {
  uma tau,m;
  
  rs[0].n = n;
  rs[0].k = k;
  rs[0].tmp = ll_vec_malloc_init(n);
  rs[0].ntmp = n;
  rs[0].gao = ll_vec_malloc_init(n + 1);
  rs[0].ngao0 = n + 1;
  rs[0].ngao = 1;
  R_zero(rs[0].gao[0]);
  vec_init(rs[0].supp);
  vec_adjust_size(rs[0].supp,n);
  
  tau = sudan_max_errors(n,k);
  rs[0].stau = tau;
  rs[0].sL = sudan_list_size(n,k,tau);
  rs[0].snx = n - tau;

  tau = guruswami_sudan_max_errors(n,k);
  m = guruswami_sudan_mult(n,k,tau);
  rs[0].gstau = tau;
  rs[0].gsL = guruswami_sudan_list_size(n,k,tau,m);
  rs[0].gsm = m;
  rs[0].gsnx = m * (n - tau);
}

/**
> `void rs_code_init_with_support(rs_code rs,uma n,uma k);`
  Initialize `rs` of length `n` and dimension `k`.
  This function must be called before doing anything else with `rs`.
  `Decoding` will attempt to create the support of `rs`:
  it will first set the first coordinate to zero, then
  it will use `R_next` to set the other coordinates.
  Make sure that the ring you are using has enough elements.
  For exemple, if the ring is a prime field `gfp` then the
  support will be set to `[0,1,2,3,...,n]`.
*/

void rs_code_init_with_support(rs_code rs,uma n,uma k) {
  uma i;
  R *supp;
  rs_code_init(rs,n,k);
  supp = rs[0].supp[0].a;
  R_zero(supp[0]);
  for( i = 1 ; i < n ; i++ ) {
    R_next(supp[i],supp[i - 1]);
  }
}

/**
> `void rs_code_clear(rs_code rs);`
  Free the memory occupied by `rs`.
*/

void rs_code_clear(rs_code rs) {
  vec_clear(rs[0].supp);
  ll_vec_free(rs[0].tmp,rs[0].ntmp);
}

/**
> `void rs_code_random(vec r,rs_code rs);`
  Set `r` to be a random codeword of `rs`.
*/

void rs_code_random_codeword(vec r,rs_code rs) {
  R *tmp;
  uma n,k;
  n = rs[0].n;
  k = rs[0].k;
  tmp = rs[0].tmp;
  ll_vec_random(tmp,k);
  vec_adjust_size(r,n);
  ll_upol_multi_eval(r[0].a,tmp,k,rs[0].supp[0].a,n);
}

/**
> `uma rs_code_sudan_koetter(vec **r,rs_code rs,vec y,uma tau);`
  Given a Reed-Solomon code `rs` and a received word `y` with at most
  `tau` errors, return the number of codewords within distance `tau`
  of `y` and set `r` to an array containing the codewords.
  The Sudan algorithm is used, the interpolation step is done with the
  Koetter algorithm.
*/

uma rs_code_sudan_koetter(vec **r,rs_code rs,vec y,uma tau) {
  uma n,k,L,tau2;
  uma nqx,nqy,nq;
  R *yptr,*q,*supp,*tmp;
  struct ll_vec_lili *roots,*v;
  vec *c;
  uma nc;
  
  /* stuff with the parameters */
  n = rs[0].n;
  k = rs[0].k;
  yptr = y[0].a;
  supp = rs[0].supp[0].a;
  if ( tau == SUDAN_MAX_ERRORS ) {
    L = rs[0].sL;
    tau2 = rs[0].stau;
  }
  else {
    L = sudan_list_size(n,k,tau);
    tau2 = tau;
  }
  nq = (sudan_list_size(n,k,tau2) + 1) * (n - tau2 + 1);
  q = ll_vec_malloc_init(nq);
  
  /* 1) interpolation step */
  if ( ll_koetter_single(q,&nqx,&nqy,supp,n,k,L,1,yptr,tau2) ) {
    return 0;
  }
  
  /* 2) root finding */
  roots = ll_dbpol_Xroots(q,nqx,nqy,k);

  /* 3) select the roots within distance tau of y */
  if ( !roots ) return 0;
  nc = 0;
  tmp = 0;
  c = 0;
  for( v = roots ; v ; v = v->next ) {
    if ( !tmp ) tmp = ll_vec_malloc_init(n);
    ll_upol_multi_eval(tmp,v->a,k,supp,n);
    if ( ll_vec_hamming_distance(yptr,tmp,n) <= tau2 ) {
      if ( nc == 0 ) {
        c = __CAST(vec*,ll_malloc(sizeof(vec)));
      }
      else {
        c = __CAST(vec*,ll_realloc(c,(nc + 1) * sizeof(vec)));
      }
      vec_init(c[nc]);
      c[nc][0].a = tmp;
      c[nc][0].na = n;
      nc++;
      tmp = 0;
    }
  }
  if ( tmp ) ll_vec_free(tmp,n);
  ll_vec_lili_list_free(roots);
  *r = c;

  /* free stuff */
  ll_vec_free(q,nq);

  return nc;
}

/**
> `uma rs_code_guruswami_sudan_koetter(vec **r,rs_code rs,vec y,uma tau);`
  Given a Reed-Solomon code `rs` and a received word `y` with at most
  `tau` errors, return the number of codewords within distance `tau`
  of `y` and set `r` to an array containing the codewords.
  The Guruswami-Sudan algorithm is used, the interpolation step is done
  with the Koetter algorithm.
*/

uma rs_code_guruswami_sudan_koetter(vec **r,rs_code rs,vec y,uma tau) {
  uma n,k,L,m,tau2;
  uma nqx,nqy,nq;
  R *yptr,*q,*supp,*tmp;
  struct ll_vec_lili *roots,*v;
  vec *c;
  uma nc;
  
  /* stuff with the parameters */
  n = rs[0].n;
  k = rs[0].k;
  yptr = y[0].a;
  supp = rs[0].supp[0].a;
  m = guruswami_sudan_mult(n,k,tau);
  if ( tau == GURUSWAMI_SUDAN_MAX_ERRORS ) {
    L = rs[0].gsL;
    tau2 = rs[0].gstau;
    m = rs[0].gsm;
  }
  else {
    m = guruswami_sudan_mult(n,k,tau);
    L = guruswami_sudan_list_size(n,k,tau,m);
    tau2 = tau;
  }
  nq = (guruswami_sudan_list_size(n,k,tau2,m) + 1) *
       (m * (n - tau2) + 1);
  q = ll_vec_malloc_init(nq);

  /* 1) interpolation step */
  if ( ll_koetter_single(q,&nqx,&nqy,supp,n,k,L,m,yptr,tau2) ) {
    return 0;
  }

  /* 2) root finding */
  roots = ll_dbpol_Xroots(q,nqx,nqy,k);

  /* 3) select the roots within distance tau of y */
  if ( !roots ) return 0;
  nc = 0;
  tmp = 0;
  c = 0;
  for( v = roots ; v ; v = v->next ) {
    if ( !tmp ) tmp = ll_vec_malloc_init(n);
    ll_upol_multi_eval(tmp,v->a,k,supp,n);
    if ( ll_vec_hamming_distance(yptr,tmp,n) <= tau2 ) {
      if ( nc == 0 ) {
        c = __CAST(vec*,ll_malloc(sizeof(vec)));
      }
      else {
        c = __CAST(vec*,ll_realloc(c,(nc + 1) * sizeof(vec)));
      }
      vec_init(c[nc]);
      c[nc][0].a = tmp;
      c[nc][0].na = n;
      nc++;
      tmp = 0;
    }
  }
  if ( tmp ) ll_vec_free(tmp,n);
  ll_vec_lili_list_free(roots);
  *r = c;

  /* free stuff */
  ll_vec_free(q,nq);

  return nc;
}

/**
> `uma rs_code_gao(vec r,rs_code rs,vec y);`
  Given a Reed-Solomon code `rs` and a received word `y`, set `r`
  to be the codeword at distance at most half the minimum distance
  of `y` if it exists, in which case `1` is returned.
  If no codeword is found `0` is returned.
  In fact `rs_code_gao` return the number of codewords around `y`.
  The Gao unique decoding algorithm is used.
*/

uma rs_code_gao(vec r,rs_code rs,vec y) {
  uma n,k,ng0,ng10,ng1,nu0,nu,nv0,nv,neea0,neea;
  uma nquo,nrem0,nrem,ret;
  R *g0,*g1,*u,*v,*eea,*quo,*rem;
  
  n = rs[0].n;
  k = rs[0].k;
  ret = 0;
  /* Pre-computation of rs[0].gao */
  if ( rs[0].ngao == 1 ) { 
    ll_upol_prod_affine_neg(rs[0].gao,rs[0].supp[0].a,rs[0].n);
    rs[0].ngao = n + 1;
  }
  
  /* interpolation */
  g0 = rs[0].gao;
  ng0 = rs[0].ngao;
  ng10 = n;
  g1 = ll_vec_malloc_init(ng10);
  ll_upol_interpolation(g1,&ng1,rs[0].supp[0].a,y[0].a,n);

  /* partial GCD */
  nu0 = n + 1;
  u = ll_vec_malloc_init(nu0);
  nv0 = n + 1;
  v = ll_vec_malloc_init(nv0);
  neea0 = n + 1;
  eea = ll_vec_malloc_init(neea0);
  ll_upol_partial_eea(eea,&neea,u,&nu,v,&nv,
                      g0,ng0,g1,ng1,(n + k) / 2);

  /* long division */
  nquo = neea - nv + 1;
  quo = ll_vec_malloc_init(nquo);
  nrem0 = neea;
  rem = ll_vec_malloc_init(nrem0);
  ll_upol_quorem(rem,&nrem,quo,eea,neea,v,nv);

  /* final stage */
  if ( ll_upol_iszero(rem,nrem) && nquo <= k ) {
    vec_adjust_size(r,n);
    ll_upol_multi_eval(r[0].a,quo,nquo,rs[0].supp[0].a,n);
    ret = 1;
  }
  
  /* free stuff */
  ll_vec_free(g1,ng10);
  ll_vec_free(u,nu0);
  ll_vec_free(v,nv0);
  ll_vec_free(eea,neea0);
  ll_vec_free(quo,nquo);
  ll_vec_free(rem,nrem0);
  return ret;
}

