/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

/* INIT, CLEAR FUNCTIONS */

/**
> `void vec_init(vec r);`
  Initialize the vector `r`.
  This function must called before anything else.
*/

void vec_init(vec_ptr r) {
  r->a = NULL;
  r->na = 0;
  r->sz = 0;
}

/**
> `void vec_inits(vec r,...);`
  Initialize a NULL-terminated list of vectors.
  This function must called before anything else.
*/

void vec_inits(vec_ptr r,...) {
  va_list ap;
  va_start(ap,r);
  
  vec_init(r);
  for(;;) {
    vec_ptr a;
    a = va_arg(ap,vec_ptr);
    if ( !a ) {
      va_end(ap);
      return;
    }
    vec_init(a);
  }
}

/**
> `void vec_clear(vec r);`
  Free the memory occupied by the vector `r`.
  You must not use `r` after a call to this function.
  If you want to use `r` you must make a call to `vec_init`.
*/

void vec_clear(vec_ptr r) {
  if ( r->a ) ll_vec_free(r->a,r->sz);
  r->a = NULL;
  r->na = 0;
  r->sz = 0;
}

/**
> `void vec_clears(vec r,...);`
  Free a NULL-terminated list of vectors.
  You must not use the vectors after a call to this function.
  If you want to use some or all of the vectors you must make
  calls `vec_init` or `vec_inits`.
*/

void vec_clears(vec_ptr r,...) {
  va_list ap;
  va_start(ap,r);
  
  vec_clear(r);
  for(;;) {
    vec_ptr a;
    a = va_arg(ap,vec_ptr);
    if ( !a ) {
      va_end(ap);
      return;
    }
    vec_clear(a);
  }
}

/* this is an internal function */

void vec_adjust_size(vec r,uma nr) {
  uma nsz;
  if ( nr <= r[0].sz ) return;
  if ( r[0].sz == 0 ) {
    r[0].a = ll_vec_malloc_init(nr);
    r[0].na = nr;
    r[0].sz = nr;
    return;
  }
  nsz = r[0].sz * 2;
  for( ; nsz < nr ; nsz *= 2 );
  r[0].a = ll_vec_realloc(r[0].a,r[0].sz,nsz);
  r[0].na = nr;
  r[0].sz = nsz;
}

/**
> `void vec_zero(vec r);`
  Set all the components of `r` to zero.
*/

void vec_zero(vec r) {
  ll_vec_zero(r[0].a,r[0].na);
}

/**
> `void vec_append(vec r,R a);`
  Append `a` to `r`.
*/
void vec_append(vec r,R a) {
  uma nnr;
  nnr = r[0].na + 1;
  vec_adjust_size(r,nnr);
  R_copy(r[0].a[nnr - 1],a);
}

/**
> `void vec_random(vec r);`
  Set all the compoenets of `r` to random elements of the ring.
*/

void vec_random(vec r) {
  ll_vec_random(r[0].a,r[0].na);
}

/**
> `void vec_random_hamming_weight(vec r,uma nr,uma w);`
  Set `r` to be a random vector of length `nr` and
  Hamming weight `w`.
*/

void vec_random_hamming_weight(vec r,uma nr,uma w) {
  if ( w > nr ) {
    ll_die("length must be grater than Hamming weight");
  }
  vec_adjust_size(r,nr);
  ll_vec_random_hamming_weight(r[0].a,r[0].na,w);
}

/* ARITHMETIC */

/**
> `void vec_add(vec r,vec a,vec b);`
  Set `r` to `a + b`.
  If `a` and `b` have not the same length
  the shortest vector is padded with zeros.
*/

void vec_add(vec r,vec a,vec b) {
  uma nnr;
  nnr = MAX(a[0].na,b[0].na);
  vec_adjust_size(r,nnr);
  ll_array_add(r[0].a,a[0].a,a[0].na,b[0].a,b[0].na);
}

/**
> `void vec_sub(vec r,vec a,vec b);`
  Set `r` to `a - b`.
  If `a` and `b` have not the same length
  the shortest vector is padded with zeros.
*/

void vec_sub(vec r,vec a,vec b) {
  uma nnr;
  nnr = MAX(a[0].na,b[0].na);
  vec_adjust_size(r,nnr);
  ll_array_sub(r[0].a,a[0].a,a[0].na,b[0].a,b[0].na);
}

/**
> `void vec_mul_by_cte(vec r,vec a,R b);`
  Set `r` to `a * b`.
*/

void vec_mul_by_cte(vec r,vec a,R b) {
  uma nnr;
  nnr = a[0].na;
  vec_adjust_size(r,nnr);
  ll_vec_mul_by_cte(r[0].a,a[0].a,nnr,b);
}

/* UTILS */

/**
> `void vec_set_coeff(vec r,R a,uma i);`
  Set the `i`th coefficient of `r` to `a`.
*/

void vec_set_coeff(vec r,R a,uma i) {
  vec_adjust_size(r,i + 1);
  R_copy(r[0].a[i],a);
}

/**
> `void vec_get_coeff(R r,vec a,uma i);`
  Set `r` to the `i`th ceofficient of `a`.
*/

void vec_get_coeff(R r,vec a,uma i) {
  if ( i < 0 || i >= a[0].na ) {
    R_zero(r);
    return;
  }
  R_copy(r,a[0].a[i]);
}

/**
> `uma vec_get_len(vec a);`
  Return the length of `a`.
*/

uma vec_get_len(vec a) {
  return a[0].na;
}

/**
> `uma vec_hamming_weight(vec a);`
  Return the Hamming weight of `a`.
*/

uma vec_hamming_weight(vec a) {
  if ( a[0].na == 0 ) return 0;
  return ll_vec_hamming_weight(a[0].a,a[0].na);
}

/**
> `uma vec_hamming_distance(vec a,vec b);`
  Return the Hamming distance between `a` and `b`.
  If `a` and `b` have the not the same size then
  `-1` is returned.
*/

uma vec_hamming_distance(vec a,vec b) {
  if ( a[0].na != b[0].na ) return -1;
  return ll_vec_hamming_distance(a[0].a,b[0].a,a[0].na);
}

/**
> `uma vec_copy(vec r,vec a);`
  Make a deep copy of `a` into `r`.
*/

void vec_copy(vec r,vec a) {
  vec_adjust_size(r,a[0].na);
  ll_vec_copy(r[0].a,a[0].a,a[0].na);
}

/**
> `uma vec_cmp(vec a,vec b);`
  Return `0` if and only if `a` equals `b`.
  If `a` and `b` do not have the same length they
  are not equal.
*/

uma vec_cmp(vec a,vec b) {
  if ( a[0].na != b[0].na ) return -1;
  return ll_vec_cmp(a[0].a,b[0].a,a[0].na);
}

/**
> `void vec_ptr_clear(vec *a,uma n);`
  Clear the array `a` of vectors of size `n`.
*/

void vec_ptr_clear(vec *a,uma n) {
  uma i;
  for( i = 0 ; i < n ; i++ ) {
    vec_clear(a[i]);
  }
  free(a);
}

/**
> `void vec_print(FILE *out,vec_ptr a);`
  Pruma `a` into `out`.
*/

void vec_print(FILE *out,vec_ptr a) {
  ll_vec_print(out,a->a,a->na);
}

