/*
 * This file is part of decoding
 * Copyright 2012  Guillaume Quintin, INRIA
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/*************************************************

           number of hardware threads

*************************************************/
#ifdef HARD_THREADS

#include <omp.h>
#include <stdio.h>
int main(void) {
  #pragma omp parallel
  printf("Hello from thread %d, nthreads %d\n",
         omp_get_thread_num(),omp_get_num_threads());
  return 0;
}
#endif

/*************************************************

            Modular floats with SSE

*************************************************/
#ifdef MOD_FLOAT_SSE
#define FP_PRIME 257.0

float reduce(float a) {
  float q;
  q = (float)((int)(a * ((float)1.0 / (float)FP_PRIME)));
  return a - q * FP_PRIME;
}

int main(int argc,char **argv) {
  int i,n;
  float *a,*b,*c;
  if ( argc != 2 ) return -1;
  n = atoi(argv[1]);
  for( i = 0 ; i < n ; i++ ) {
    a[i] = reduce(b[i] + c[i]);
  }
  printf("%f\n",a[rand() % n]);
  return 0;
}
#endif

/*************************************************

         Modular arithmetic with floats

*************************************************/
#ifdef MOD_FLOAT
int main(void) {
  int i,n;
  for( i = 3 ; i < (1 << 20) ; i++ ) {
    float a,b;
    a = (float)(i * i - 1);
    b = (float)i;
    if ( (int)(a / b) != (i - 1) ) {
      fprintf(stderr,"Error: i=%d, a/b=%f\n",i,a / b);
      exit(0);
    }
  }
}
#endif


/*************************************************

         Aligned VS unaligned mem access

*************************************************/
#ifdef ALIGN_VS_UNALIGN
void *malloc2(int n) {
  void *p;
  char *q;
  unsigned long a,b;
  p = malloc(n + 16);
  if ( !p ) exit(-1);
  a = (unsigned long)p;
  b = a % 16;
  if ( b == 0 ) return p;
  q = p;
  q += 16 - b;
  q[-1] = (char)(16 - b);
  return q;
}

int main(int argc,char **argv) {
  int *a,*b,c;
  unsigned long n,i;
  time_t start,end;

  if ( argc != 2 ) return -1;
  n = atol(argv[1]);

  printf("%d: ",n); fflush(stdout);

#ifdef ALIGN
  a = __builtin_assume_aligned(malloc2(n * sizeof(int)),16);
  b = __builtin_assume_aligned(malloc2(n * sizeof(int)),16);
#else
  a = malloc(n * sizeof(int));
  b = malloc(n * sizeof(int));
  if ( !a || !b ) return -1;
#endif

  srand(time(0));

  for( i = 0 ; i < n ; i++ ) {
    a[i] = rand();
  }
  c = rand();

  start = clock();
  for( i = 1 ; i < n - 1 ; i++ ) {
    b[i] = a[i] + 0x12AB34;
  }
  end = clock();

  printf("%f\n",((double)end - (double)start) / CLOCKS_PER_SEC);

  return b[rand() % n];
}
#endif

