/*
 * This file is part of decoding
 * Copyright 2013  Guillaume Quintin
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#define VERSION "cpuid version 0.2"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* Intel stuff
   http://download.intel.com/products/processor/manual/325383.pdf */
#define INTEL_STRING_B   0x756E6547u
#define INTEL_STRING_C   0x6C65746Eu
#define INTEL_STRING_D   0x49656E69u
#define INTEL_AVX        (1u << 28)
#define INTEL_SSE        (1u << 25)
#define INTEL_SSE2       (1u << 26)
#define INTEL_SSE3       (1u << 0)
#define INTEL_SSSE3      (1u << 9)
#define INTEL_SSE41      (1u << 19)
#define INTEL_SSE42      (1u << 20)
#define INTEL_AES        (1u << 25)
#define INTEL_PCLMULQDQ  (1u << 1)
#define INTEL_HTT        (1u << 28)

/* AMD stuff
   http://support.amd.com/us/Processor_TechDocs/25481.pdf */
#define AMD_STRING_B     0x68747541u
#define AMD_STRING_C     0x444D4163u
#define AMD_STRING_D     0x69746E65u
#define AMD_AVX          (1u << 28)
#define AMD_SSE          (1u << 25)
#define AMD_SSE2         (1u << 26)
#define AMD_SSE3         (1u << 0)
#define AMD_SSSE3        (1u << 9)
#define AMD_SSE41        (1u << 19)
#define AMD_SSE42        (1u << 20)
#define AMD_AES          (1u << 25)
#define AMD_PCLMULQDQ    (1u << 1)
#define AMD_HTT          (1u << 28)

/* My stuff */
#define NONE             0u
#define SSE              (1u << 0)
#define SSE2             (1u << 1)
#define SSE3             (1u << 2)
#define SSSE3            (1u << 3)
#define SSE41            (1u << 4)
#define SSE42            (1u << 5)
#define AVX              (1u << 0)
#define PCLMULQDQ        (1u << 1)
#define AES              (1u << 2)
#define HTT              (1u << 3)

typedef struct cpu cpu;
struct cpu {
  unsigned int sse;
  unsigned int other;
  unsigned int logical_cores;
  unsigned int physical_cores;
};

/* Call to the cpuid instruction if supported,
   otherwise (0,0,0,0) is returned */
void cpuid(unsigned int eax,unsigned int ecx,
           unsigned int *a,unsigned int *b,
           unsigned int *c,unsigned int *d)
{
#if ( __GNUC__ || __clang__ ) && ( __x86_64__ || __i386__ )
#ifdef __GNUC__
  __asm__ __volatile__ ("cpuid": \
  "=a" (*a) , "=b" (*b) , "=c" (*c) , "=d" (*d) : \
  "a" (eax) , "c" (ecx) );
#endif
#ifdef __clang__
  asm ("cpuid": \
  "=a" (*a) , "=b" (*b) , "=c" (*c) , "=d" (*d) : \
  "a" (eax) , "c" (ecx) );
#endif
#else
  (void)eax;
  a[0] = 0;
  b[0] = 0;
  c[0] = 0;
  d[0] = 0;
#endif
}

/* Intel processor */
void intel_cpuid(cpu *buf) {
  unsigned int a,b,c,d;
  cpuid(1u,0u,&a,&b,&c,&d);
  buf->sse = NONE;
  if ( d & INTEL_SSE ) buf->sse |= SSE;
  if ( d & INTEL_SSE2 ) buf->sse |= SSE2;
  if ( c & INTEL_SSE3 ) buf->sse |= SSE3;
  if ( c & INTEL_SSSE3 ) buf->sse |= SSSE3;
  if ( c & INTEL_SSE41 ) buf->sse |= SSE41;
  if ( c & INTEL_SSE42 ) buf->sse |= SSE42;
  buf->other = NONE;
  if ( c & INTEL_AVX ) buf->other |= AVX;
  if ( c & INTEL_AES ) buf->other |= AES;
  if ( c & INTEL_PCLMULQDQ ) buf->other |= PCLMULQDQ;
  if ( d & INTEL_HTT ) {
    buf->other |= HTT;
    buf->logical_cores = ((b >> 16) & 0xFFu);
    cpuid(4u,0u,&a,&b,&c,&d);
    buf->physical_cores = ((a >> 26) & 0x3Fu) + 1;
  }
  else {
    buf->logical_cores = 1;
    buf->physical_cores = 1;
  }
}

/* AMD processor: WARNING => NOT TESTED <= */
void amd_cpuid(cpu *buf) {
  unsigned int a,b,c,d;
  cpuid(1u,0u,&a,&b,&c,&d);
  buf->sse = NONE;
  if ( d & AMD_SSE ) buf->sse |= SSE;
  if ( d & AMD_SSE2 ) buf->sse |= SSE2;
  if ( c & AMD_SSE3 ) buf->sse |= SSE3;
  if ( c & AMD_SSSE3 ) buf->sse |= SSSE3;
  if ( c & AMD_SSE41 ) buf->sse |= SSE41;
  if ( c & AMD_SSE42 ) buf->sse |= SSE42;
  buf->other = NONE;
  if ( c & AMD_AVX ) buf->other |= AVX;
  if ( c & AMD_AES ) buf->other |= AES;
  if ( c & AMD_PCLMULQDQ ) buf->other |= PCLMULQDQ;
  if ( d & AMD_HTT ) {
    buf->other |= HTT;
    buf->logical_cores = ((b >> 16) & 0xFFu);
    cpuid(0x80000008u,0u,&a,&b,&c,&d);
    buf->physical_cores = c & 0xFFu;
  }
  else {
    buf->logical_cores = 1;
    buf->physical_cores = 1;
  }
}

void print_version(FILE *out) {
  fputs(VERSION "\n",out);
}

void print_usage(FILE *out) {
  print_version(out);
  fputs("usage: cpuid [-v | -f | -h prefix]\n\n"
        "  Print the list of capabilities of your processor\n\n"
        "  -v  Print the version of cpuid.\n"
        "  -f  Print the compile flags to give gcc\n"
        "      instead of the list of capabilities.\n"
        "  -h  Print C defines corresponding to the capabilities\n"
        "      of your processor. All macros are prefixed by\n"
        "      ``prefix\"\n",out);
}

void fill_cpu(cpu *buf) {
  unsigned int a,b,c,d;
  cpuid(0u,0u,&a,&b,&c,&d);
  if ( b == INTEL_STRING_B &&
       c == INTEL_STRING_C &&
       d == INTEL_STRING_D )
  {
    intel_cpuid(buf);
    return;
  }
  if ( b == AMD_STRING_B &&
       c == AMD_STRING_C &&
       d == AMD_STRING_D )
  {
    amd_cpuid(buf);
    return;
  }
}

void print_header(FILE *out,cpu *buf,const char *prefix) {
  if ( buf->sse & SSE ) fprintf(out,"#define %sSSE\n",prefix);
  if ( buf->sse & SSE2 ) fprintf(out,"#define %sSSE2\n",prefix);
  if ( buf->sse & SSE3 ) fprintf(out,"#define %sSSE3\n",prefix);
  if ( buf->sse & SSSE3 ) fprintf(out,"#define %sSSSE3\n",prefix);
  if ( buf->sse & SSE41 ) fprintf(out,"#define %sSSE41\n",prefix);
  if ( buf->sse & SSE42 ) fprintf(out,"#define %sSSE42\n",prefix);
  if ( buf->other & AVX ) fprintf(out,"#define %sAVX\n",prefix);
  if ( buf->other & PCLMULQDQ ) {
    fprintf(out,"#define %sPCLMULQDQ\n",prefix);
  }
  if ( buf->other & AES ) fprintf(out,"#define %sAES\n",prefix);
  fprintf(out,"#define %sPHYSICAL_CORES %u\n",
          prefix,buf->physical_cores);
  fprintf(out,"#define %sLOGICAL_CORES %u\n",
          prefix,buf->logical_cores);
}

void print_capabilities(FILE *out,cpu *buf) {
  if ( buf->sse & SSE ) fputs("SSE\n",out);
  if ( buf->sse & SSE2 ) fputs("SSE2\n",out);
  if ( buf->sse & SSE3 ) fputs("SSE3\n",out);
  if ( buf->sse & SSSE3 ) fputs("SSSE3\n",out);
  if ( buf->sse & SSE41 ) fputs("SSE41\n",out);
  if ( buf->sse & SSE42 ) fputs("SSE42\n",out);
  if ( buf->other & AVX ) fputs("AVX\n",out);
  if ( buf->other & PCLMULQDQ ) fputs("PCLMULQDQ\n",out);
  if ( buf->other & AES ) fputs("AES\n",out);
  printf("PHY_CORES=%u\nLOGI_CORES=%u\n",
         buf->physical_cores,buf->logical_cores);
}

void print_gcc_flags(FILE *out,cpu *buf) {
  int sp;
  sp = 0;
  if ( buf->sse & SSE42 ) {
    if ( sp ) fputc(' ',out);
    fputs("-msse4.2",out);
    sp = 1;
  }
  else {
    if ( buf->sse & SSE41 ) {
      if ( sp ) fputc(' ',out);
      fputs("-msse4.1",out);
      sp = 1;
    }
    else {
      if ( buf->sse & SSSE3 ) {
        if ( sp ) fputc(' ',out);
        fputs("-mssse3",out);
        sp = 1;
      }
      else {
        if ( buf->sse & SSE3 ) {
          if ( sp ) fputc(' ',out);
          fputs("-msse3",out);
          sp = 1;
        }
        else {
          if ( buf->sse & SSE2 ) {
            if ( sp ) fputc(' ',out);
            fputs("-msse2",out);
            sp = 1;
          }
          else {
            if ( buf->sse & SSE ) {
              if ( sp ) fputc(' ',out);
              fputs("-msse",out);
              sp = 1;
            }
          }
        }
      }
    }
  }
  if ( buf->other & AVX ) {
    if ( sp ) fputc(' ',out);
    fputs("-mavx",out);
    sp = 1;
  }
  if ( buf->other & PCLMULQDQ ) {
    if ( sp ) fputc(' ',out);
    fputs("-mpclmul",out);
    sp = 1;
  }
  if ( buf->other & AES ) {
    if ( sp ) fputc(' ',out);
    fputs("-maes",out);
    sp = 1;
  }
  fputc('\n',out);
}

int main(int argc,char **argv) {
  cpu buf;
  if ( argc == 2 && !strcmp(argv[1],"-v") ) {
    print_version(stdout);
    return 0;
  }
  if ( argc > 3 ) {
    print_usage(stderr);
    exit(-1);
  }
  fill_cpu(&buf);
  if ( argc == 1 ) {
    print_capabilities(stdout,&buf);
    return 0;
  }
  if ( argc == 2 && !strcmp(argv[1],"-f") ) {
    print_gcc_flags(stdout,&buf);
    return 0;
  }
  if ( argc == 3 && !strcmp(argv[1],"-h") ) {
    print_header(stdout,&buf,argv[2]);
    return 0;
  }
  print_usage(stderr);
  return -1;
}

