/*
 * This file is part of decoding
 * Copyright 2012-2013  Guillaume Quintin
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */

#define _POSIX_C_SOURCE 1

#define VERSION "sizeof version 0.2"

#define ERROR  -1
#define CHAR    0
#define SHORT   1
#define INT     2
#define LONG    3
#define FLOAT   4
#define DOUBLE  5

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <limits.h>

void print_version(FILE *out) {
  fputs(VERSION "\n",out);
}

void print_usage(FILE *out) {
  print_version(out);
  fputs("usage: sizeof [-v | [-b] type]\n\n"
        "  Print the size in bytes of the basic C types.\n\n"
        "  -v  Print the version of sizeof.\n"
        "  -b  Print the size in bits of the type.\n"
        "  type\n"
        "      Available C types are signed and unsigned\n"
        "      char, short, int, long, float, double, void*,\n"
        "      char*, short*, int*, long*, float* and double*.\n",
        out);
}

int get_type(const char *type) {
  if ( strstr(type,"char") ) return CHAR;
  if ( strstr(type,"short") ) return SHORT;
  if ( strstr(type,"int") ) return INT;
  if ( strstr(type,"long") ) return LONG;
  if ( strstr(type,"float") ) return FLOAT;
  if ( strstr(type,"double") ) return DOUBLE;
  return ERROR;
}

int is_signed(const char *type) {
  if ( strstr(type,"unsigned") ) return -1;
  return 0;
}

int is_pointer(const char *type) {
  if ( strstr(type,"*") ) return -1;
  return 0;
}

#define HELPER(type) \
  switch(sel) { \
    case 0: /* unsigned */ \
      sz = sizeof(unsigned type); \
      break; \
    case 1: /* signed */ \
      sz = sizeof(signed type); \
      break; \
    case 2: /* unsigned pointer */ \
      sz = sizeof(unsigned type*); \
      break; \
    case 3: /* signed pointer */ \
      sz = sizeof(signed type*); \
      break; \
  }

int get_sizeof(const char *type,int mult) {
  int sgn,ptr,typ,sel,sz;
  sgn = ( is_signed(type) ? 1 : 0 );
  ptr = ( is_pointer(type) ? 1 : 0 );
  sel = (ptr << 1) | sgn;
  typ = get_type(type);
  switch(typ) {
    case CHAR:
      HELPER(char)
      break;
    case SHORT:
      HELPER(short)
      break;
    case INT:
      HELPER(int)
      break;
    case LONG:
      HELPER(long)
      break;
    case FLOAT:
      sz = ( ptr ? sizeof(float*) : sizeof(float) );
      break;
    case DOUBLE:
      sz = ( ptr ? sizeof(double*) : sizeof(double) );
      break;
    case ERROR:
    default:
      fprintf(stderr,"sizeof: unknown type\n\n");
      print_usage(stderr);
      exit(-1);
  }
  return sz * mult;
}

#undef HELPER

int main(int argc,char **argv) {
  int mult,id;
  if ( argc == 2 && !strcmp(argv[1],"-v") ) {
    fputs(VERSION "\n",stdout);
    exit(0);
  }
  if ( argc != 2 && argc != 3 ) {
    print_usage(stderr);
    exit(-1);
  }
  if ( argc == 3 && strcmp(argv[1],"-b") ) {
    print_usage(stderr);
    exit(-1);
  }
  if ( argc == 2 ) {
    mult = 1;
    id = 1;
  }
  else if ( argc == 3 ) {
    mult = CHAR_BIT;
    id = 2;
  }
  fprintf(stdout,"%d\n",get_sizeof(argv[id],mult));
  fflush(stdout);
  return 0;
}

